"""
File Purpose: Behavior object
- like a dict, but with extra features.
- for describing the current behavior of a PlasmaCalculator
"""

import itertools
import sys

import numpy as np
import xarray as xr

from ..errors import UnitsError, InputConflictError
from ..tools import (
    simple_property, alias,
    UNSET, ATTR_UNSET,
    repr_simple,
    format_docstring,
    xarray_assign, is_iterable_dim,
)
from ..defaults import DEFAULTS


class Behavior(dict):
    '''dict with 'compatible' method.

    __getitem__ altered to include values from self.dims, and 'dims' itself.
        e.g. self['dims'] == self.dims; self['fluid'] --> self.dims['fluid'], if possible.
    __contains__ altered to include values from self.dims.
        e.g. 'fluid' in self --> fluid in self.keys() or fluid in self.dims

    arg0: object
        first argument to pass to dict() constructor.
    dims: None or dict
        dict of dimension values for this behavior.
        if None, use empty dict.
    remaining **kw are passd to dict() constructor
    '''
    # # # CREATION / INITIALIZATION # # #
    def __init__(self, arg0=UNSET, dims=None, **kw):
        '''initialize Behavior.'''
        if arg0 is UNSET:
            super().__init__(**kw)
        else:
            super().__init__(arg0, **kw)
        self.dims = dict() if dims is None else dims

    def _new_with_dims(self, dims):
        '''return new Behavior like self but with dims instead of self.dims.'''
        return type(self)(super().copy(), dims=dims)

    # # # DICT-LIKE FOR 'dims' # # #
    def __getitem__(self, key):
        '''return self[key], or self.dims[key] if key is in self.dims.
        if key is 'dims', return self.dims.
        '''
        if key == 'dims':
            return self.dims
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.dims[key]

    def __contains__(self, key):
        '''return whether key is in self or self.dims.'''
        return super().__contains__(key) or key in self.dims

    def copy(self, *, keys=None):
        '''return a copy of self, including a copy of self.dims.
        keys: None or iterable
            if provided, the copy will include keys only from this list.
            (either way, the copy will include a full copy of self.dims.)
        '''
        copy_dims = self.dims.copy()
        if keys is None:
            copy_dict = super().copy()
        else:
            copy_dict = {k: self[k] for k in keys}
        return type(self)(copy_dict, dims=copy_dims)

    # # # COMPARISON # # #
    def compatible(self, other, *, lenient=False, subdims=False):
        '''return whether self and other are compatible.
        'compatible' when self[key] == other[key] for all keys in self (and self.dims).
        only tests keys in self; other can have more keys than those in self.

        lenient: bool
            whether to only test keys which are in both self and other.
            if False, and other is missing any keys in self, return False.
        subdims: bool
            whether to be compatible if dim from self is a value within dim from other.
            when True, if only compatible due to subdims, return list of dims which are subdims.
            E.g., self = dict(fluid=1, snap=0), other = dict(fluid=[0,1,2,3], snap=0) -->
                subdims=True --> compatible; result will be ['fluid']
                subdims=False --> not compatible.
            To check whether other dim is a "list" of values, use is_iterable_dim(other_val)
        '''
        if not lenient:
            if not all(key in other for key in (*self, *self.dims)):
                return False  # other is missing key; cannot be compatible.
        # non-dims compatibility
        for key, val in self.items():
            try:
                other_val = other[key]
            except KeyError:  # skip this key, if lenient.
                assert lenient, f"coding error... other is missing {key!r}; should've already returned False"
            else:
                if val != other_val:
                    return False
        # dims compatibility
        subdims_result = []
        for key, val in self.dims.items():
            try:
                other_val = other[key]
            except KeyError:  # skip this key, if lenient.
                assert lenient, f"coding error... other is missing {key!r}; should've already returned False"
            else:
                if val != other_val:
                    if subdims and is_iterable_dim(other_val) and (val in other_val):
                        subdims_result.append(key)
                    else:
                        return False
        return True if len(subdims_result)==0 else subdims_result

    # # # CONVENIENCE # # #
    @format_docstring(DEFAULTS=DEFAULTS)
    def label_array(self, array, *, skip=UNSET):
        '''label array with coords from self.dims, attrs from self.items().
        Requires that values of self.dims are all non-iterable, according to is_iterable_dim().

        skip: None, UNSET, or list of strs
            skip these attrs, if provided.
            UNSET --> skip DEFAULTS.SKIP_BEHAVIOR_ARRAY_LABELS
                        (default {DEFAULTS.SKIP_BEHAVIOR_ARRAY_LABELS})
            None --> don't skip anything.
        '''
        coords = self.dims
        if skip is UNSET:
            skip = DEFAULTS.SKIP_BEHAVIOR_ARRAY_LABELS
        elif skip is None:
            skip = []
        attrs = {k: v for k, v in self.items() if k not in skip}
        return xarray_assign(array, coords=coords, attrs=attrs)

    def assign_attrs(self, array):
        '''return copy of array with self keys & vals (but not dims) assigned as attrs.
        if any were already attrs of array, crash if value mismatch.

        See also: assign_nondefault_attrs
        '''
        to_assign = self
        existing = array.attrs
        for e, v in existing.items():
            if e in to_assign and to_assign[e] != v:
                errmsg = ('assign_nondefault_attrs refusing to overwrite existing attr: '
                          f'{e!r} from old value {v!r} to new value {to_assign[e]}')
                raise QuantCalcError(errmsg)
        return array.assign_attrs(to_assign)

    def assign_nondefault_attrs(self, array, ql=None, *, include_xr=True):
        '''return copy of array with self.nondefault() keys & vals (but not dims) assigned as attrs.
        if any were already attrs of array, crash if value mismatch.

        ql: None or BehaviorHaver
            if provided, used to evaluate some defaults (anything defined with getdefault)
            e.g. for stat_dims, getdefault = lambda ql: getattr(ql, 'maindims', []).
        include_xr: bool
            whether to include keys whose values are xarray objects (DataArray or Dataset).

        See also: assign_attrs
        '''
        # [TODO] reduce repeated code from above
        to_assign = self.nondefault(ql=ql, include_xr=include_xr)
        existing = array.attrs
        for e, v in existing.items():
            if e in to_assign and to_assign[e] != v:
                errmsg = ('assign_nondefault_attrs refusing to overwrite existing attr: '
                          f'{e!r} from old value {v!r} to new value {to_assign[e]}')
                raise QuantCalcError(errmsg)
        return array.assign_attrs(to_assign)

    def default_keys(self, ql=None):
        '''return list of keys in self with key.is_default(value, ql=ql)
        Does not consider self.dims.
        ql: None or BehaviorHaver
            if provided, used to evaluate some defaults (anything defined with getdefault)
            e.g. for stat_dims, getdefault = lambda ql: getattr(ql, 'maindims', []).
        '''
        result = []
        for k, v in self.items():
            if hasattr(k, 'is_default') and k.is_default(v, ql=ql):
                result.append(k)
        return result

    def nondefault_keys(self, ql=None, *, include_xr=True):
        '''return list of keys in self with NOT key.is_default(value, ql=ql).
        Does not consider self.dims.
        ql: None or BehaviorHaver
            if provided, used to evaluate some defaults (anything defined with getdefault)
            e.g. for stat_dims, getdefault = lambda ql: getattr(ql, 'maindims', []).
        include_xr: bool or 'only'
            whether to include keys whose values are xarray objects (DataArray or Dataset).
            'only' --> instead return only keys whose values are xarray objects (& not default).
        '''
        result = []
        for k, v in self.items():
            if not (hasattr(k, 'is_default') and k.is_default(v, ql=ql)):
                if include_xr == 'only':
                    if isinstance(v, (xr.DataArray, xr.Dataset)):
                        result.append(k)
                elif include_xr or not isinstance(v, (xr.DataArray, xr.Dataset)):
                    result.append(k)
        return result

    def nondefault(self, ql=None, *, include_xr=True):
        '''returns Behavior like self but dropping all keys with key.is_default(value, ql=ql).
        Note: these are the keys which MIGHT not have default values;
            result includes keys where default was unknown and unchecked.

        ql: None or BehaviorHaver
            if provided, used to evaluate some defaults (anything defined with getdefault)
            e.g. for stat_dims, getdefault = lambda ql: getattr(ql, 'maindims', []).
        include_xr: bool
            whether to include keys whose values are xarray objects (DataArray or Dataset).
            'only' --> instead return a dict of only keys whose values are xarray objects (& not default).
        '''
        if include_xr == 'only':
            return {k: self[k] for k in self.nondefault_keys(ql=ql, include_xr='only')}
        else:
            return self.copy(keys=self.nondefault_keys(ql=ql, include_xr=include_xr))

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'{k}={repr_simple(v)}' for k, v in self.items()]
        contents.append(f'dims={repr_simple(self.dims)}')
        return f'{type(self).__name__}({", ".join(contents)})'

    def __str__(self):
        contents = [f'{k}={v!r}' for k, v in self.items()]
        contents.append(f'dims={self.dims}')
        return f"{type(self).__name__}({', '.join(contents)})"


class BehaviorQuantity():
    '''quantity, with value and associated behavior.
    behavior: dict or Behavior
        this BehaviorQuantity is relevant only when behavior.compatible(obj.behavior).
        E.g., behavior = dict(snap=0, units='si'),
            - relevant if obj.behavior = dict(snap=0, units='si', ...).
            - irrelevant if obj.behavior = dict(snap=1, ...).
            - irrelevant if obj.behavior = dict(units='raw', ...)
        note: if ukey is not None, and behavior is a Behavior,
            self will store a COPY of behavior, not the original object;
            this is to prevent popping 'units' from the original Behavior object.

    MB: None or number
        number of megabytes this BehaviorQuantity takes up in memory.
        if None, will be calculated from self.value, if possible.
    ukey: None, or str
        if provided (not None), pop 'units' from stored behavior, and store self.uinfo = (ukey, units).
        None --> don't pop 'units' from stored behavior.
        str --> value [any unit system] = self.value * u_convert, where
                    u_convert = UnitsHaver.u(self.uinfo[0], (that unit system), convert_from=self.uinfo[0])
                Note, for a dimensionless value, use ukey='' or ukey='1'.
        The benefit of using this is that 'units' won't be in self.behavior,
            so self.compatible(other_behavior) won't need to check units.
        (The downside is that you must be careful to properly handle the units. see also: self.get_value.)

    Every BehaviorQuantity has a unique identifier, self.id, which is a number.
    BehaviorQuantity.NEXT_UNIQUE_ID is the next number to be assigned.
    '''
    NEXT_UNIQUE_ID = 0

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, value, behavior, *, MB=None, ukey=None, **kw__behavior):
        self.value = value
        if not isinstance(behavior, Behavior):
            behavior = Behavior(behavior, **kw__behavior)
        if ukey is not None:
            behavior = behavior.copy()
            self.uinfo = (ukey, behavior.pop('units', None))
        self.behavior = behavior
        self._entered_MB = MB
        self.n_used = 0  # counts number of times this value was loaded from cache.
                         # not updated by this class, but see e.g. VarCache.
        self.id = self.NEXT_UNIQUE_ID
        type(self).NEXT_UNIQUE_ID = self.id + 1  # [TODO] is this thread-safe?

    def _new_with_dims(self, dims, **kw_init):
        '''return new BehaviorQuantity like self but with dims instead of self.behavior.dims.'''
        kw_init.setdefault('MB', self.MB)
        return type(self)(self.value, self.behavior._new_with_dims(dims), **kw_init)

    # # # PROPERTIES # # #
    uinfo = simple_property('_uinfo', default=None,
            doc='''(ukey, units), indicating units of self.value;
            self.value [any unit system] = self.value * u_convert, where
                u_convert = UnitsHaver.u(ukey, (that unit system), convert_from=ukey)''')

    # # # GETTING VALUE # # #
    def get_value(self, units_manager=None):
        '''return self.value; if units_manager is provided return value in units_manager.units unit system.
        if self.uinfo[1] != units_manager.units, and value is an xarray with 'units' attr,
            result will have units attr assigned to units_manager.units.

        raise UnitsError if ALL of the following are True:
            - units_manager was provided
            - self.uinfo is None
            - units_manager.units!=self.behavior['units']
        '''
        u = units_manager
        if u is None:
            return self.value
        if self.uinfo is None:
            if self.behavior.get('units', None) != u.units:
                errmsg = (f"can't get value in units={u.units!r}, since self.uinfo is None, "
                          f"and self.behavior['units'] != {u.units!r}.")
                raise UnitsError(errmsg)
            else:
                return self.value
        else:
            value = self.value
            ukey, value_units = self.uinfo
            if value_units == u.units:
                result = value
            else:
                u_convert = u(ukey, convert_from=value_units)
                result = self.value * u_convert
                if isinstance(result, xr.DataArray) and result.attrs.get('units', None) != u.units:
                    result = result.assign_attrs(units=u.units)
            return result
        assert False, 'coding error if reached this line'

    # # # MATCHING OTHER BEHAVIORS # # #
    def compatible(self, behavior, *, lenient=False, subdims=False):
        '''tells whether this cached quantity is compatible to obj with this behavior.
        returns self.behavior.compatible(behavior, lenient=lenient, subdims=subdims)
        '''
        return self.behavior.compatible(behavior, lenient=lenient, subdims=subdims)

    def matches_behavior(self, behavior):
        '''tells whether this cached quantity matches this behavior.
        returns self.compatible(behavior, lenient=True, subdims=False).
        '''
        return self.compatible(behavior, lenient=True, subdims=False)

    def relevant(self, behavior):
        '''tells whether this cached quantity is relevant to this behavior.

        'relevant' when self[key] == other[key] for all keys in self,
            and self.dims[d] equals or is inside other.dims[d], for all d in self.dims.
        only tests keys in self; other can have more keys than those in self.

        returns self.compatible(behavior, lenient=True, subdims=True).
        '''
        return self.compatible(behavior, lenient=True, subdims=True)

    # # # LISTING POINTS # # #
    def list_points(self):
        '''returns a list of CachedQuantity objects, each with a single point (in dimension space) from self.
        dimension space includes dimensions but not maindims, e.g. fluid, snap, but not x,y,z.
        '''
        behavior = self.behavior
        keys, vals = zip(*behavior.dims.items())
        iterable = [is_iterable_dim(v, min_length=1) for v in vals]
        if not any(iterable):
            return [self]
        iters = [iter(v) if it_ else [v] for v, it_ in zip(vals, iterable)]
        # [TODO] <-- re-use the iter_dim_multi code from dimension_tools?
        result = []
        for vs_ in itertools.product(*iters):
            dims_point = dict(zip(keys, vs_))
            result.append(self._new_with_dims(dims_point))
            # [TODO] ^ that just copy-pastes the value from self, to each of the points.
            #    if the value is an xarray with dimension names as coords,
            #    then we should instead iterate over those coords too.
            #    e.g. if value has fluid=[0,1,2] and self.behavior.dims has fluid=[0,1,2],
            #      then the result should have 3 points, first with fluid 0, then fluid 1, then 2.
            #      right now, the result would be 3 points, each with fluid=[0,1,2].
        return result

    # # # MB # # #
    @property
    def MB(self):
        '''number of megabytes this cached quantity takes up in memory.
        if self._entered_MB is None, will be calculated from self.value.
        if self.value.nbytes is not available, will be 0.
        Saves value to self._MB to avoid recalculating.
        '''
        try:
            return self._MB
        except AttributeError:
            pass  # this is the first time getting MB; handle it below.
        if self._entered_MB is None:
            try:
                nbytes = self.value.nbytes
            except AttributeError:
                nbytes = sys.getsizeof(self.value)
            self._MB = nbytes / 1024**2
        else:
            self._MB = self._entered_MB
        return self._MB

    # # # DISPLAY # # #
    def __repr__(self):
        contents = self._repr_contents()
        return f'{type(self).__name__}({", ".join(contents)})'

    def _repr_contents(self):
        '''return list of contents to put in repr of self.'''
        val_str = repr(self.value)
        if len(val_str) > 20:
            val_str = val_str[:20] + '...'
        contents = [val_str, str(self.behavior)]
        if self.MB is not None:
            contents.append(f'MB={self.MB:.2e}')
        if self.uinfo is not None:
            contents.append(f'uinfo={self.uinfo}')
        return contents


class BehaviorAttr(str):
    '''string which refers to an attr controlling behavior of a class.

    might also know cls_where_defined, and is_dimension.

    default: ATTR_UNSET or any value
        if provided, does NOT affect the attribute itself;
        it's just a reference point for "maybe we don't need to show the info if == default"
    getdefault: ATTR_UNSET or callable of 1 arg
        similar to default but callable of BehaviorHaver: default=getdefault(behavior_haver),
            instead of being the same value for all BehaviorHavers.
        cannot provide both default and getdefault
    '''
    def __new__(cls, value, *, cls_where_defined=None, is_dimension=False, doc=None,
                default=ATTR_UNSET, getdefault=ATTR_UNSET):
        '''create new BehaviorAttr.
        cls_where_defined: None or type
            class where this attr is defined.
        is_dimension: bool
            whether this attr is a dimension.
        doc: None or str
            arbitrary docstring describing this attr.
        default: ATTR_UNSET or any object
            if provided (i.e. not ATTR_UNSET), tells behavior management code the "default" value.
            E.g. calculator.behavior.nondefault() includes only the non-default values.
            Does NOT affect the attribute itself (e.g. doesn't mess with fget & fset if property).
            Incompatible with is_dimension=True.
        getdefault: ATTR_UNSET or callable of 1 arg
            similar to default but as a callable of BehaviorHaver. See help(BehaviorAttr) for details.
        '''
        self = super().__new__(cls, value)
        self.cls_where_defined = cls_where_defined
        self.is_dimension = is_dimension
        self.doc = doc
        self.default = default
        self.getdefault = getdefault
        if is_dimension:
            if (default is not ATTR_UNSET):
                raise InputConflictError(f'cannot provide default when is_dimension. Got {default!r}')
            if (getdefault is not ATTR_UNSET):
                raise InputConflictError(f'cannot provide getdefault when is_dimension. Got {getdefault!r}')
        if (default is not ATTR_UNSET) and (getdefault is not ATTR_UNSET):
            raise InputConflictError('cannot provide both default and getdefault.')
        return self

    cls = alias('cls_where_defined')

    def is_default(self, value, ql=None):
        '''returns whether value == self.default.
        if no default was set, return False without checking equality.
        if set self.getdefault instead of default, call as getdefault(ql).
            (if ql None, return False without checking equality.)

        ql: None or BehaviorHaver
            if provided, used to evaluate some defaults (anything defined with getdefault)
            e.g. for stat_dims, getdefault = lambda ql: getattr(ql, 'maindims', []).
        '''
        default = self.default
        if default is ATTR_UNSET:
            if self.getdefault is ATTR_UNSET:
                return False
            elif ql is None:
                return False
            else:
                default = self.getdefault(ql)
        ndim = getattr(value, 'ndim', None)
        if (ndim is not None) and ndim >= 1:
            return np.all(value == default)
        else:
            return (value == default)

    def __repr__(self):
        return self._pretty_repr(fmt=None)

    def _pretty_repr(self, fmt=None):
        '''repr but align use fmt.format(super().__repr__()), if provided.'''
        contents = []
        if fmt is None:
            contents.append(super().__repr__())
        else:
            contents.append(fmt.format(super().__repr__()))
        if self.cls is not None:
            contents.append(f'cls={self.cls.__name__}')
        if self.is_dimension:
            contents.append(f'is_dimension={self.is_dimension}')
        if self.default is not ATTR_UNSET:
            contents.append(f'default={self.default}')
        return f'{type(self).__name__}({", ".join(contents)})'


class ClsBehaviorAttrs(list):
    '''list of attrs which control behavior of a class.
    also provides a method to register a new behavior attr.

    might also know cls_associated_with.
    '''
    behavior_attr_cls = BehaviorAttr

    cls_associated_with = simple_property('_cls_associated_with', default=None,
            doc='''class with which this list of behavior attrs is associated.''')

    def register(self, *attrs, is_dimension=False, default=ATTR_UNSET, getdefault=ATTR_UNSET):
        '''register attrs as BehaviorAttrs for self.cls_associated_with.
        returns list of newly registered BehaviorAttr objects.

        default: ATTR_UNSET or any object
            if provided (i.e. not ATTR_UNSET), tells behavior management code the "default" value.
            E.g. calculator.behavior.nondefault() includes only the non-default values.
            Does NOT affect the attribute itself (e.g. doesn't mess with fget & fset if property).
        getdefault: ATTR_UNSET or callable of 1 arg
            similar to default but callable of BehaviorHaver: default=getdefault(behavior_haver),
                instead of being the same value for all BehaviorHavers.
            cannot provide both default and getdefault
        '''
        cls = self.cls_associated_with
        result = []
        for attr in attrs:
            if not isinstance(attr, self.behavior_attr_cls):
                attr = self.behavior_attr_cls(attr, cls_where_defined=cls, is_dimension=is_dimension,
                                              default=default, getdefault=getdefault)
            result.append(attr)
        self.extend(result)
        return result

    def __repr__(self):
        attrs = sorted(self)
        if len(self) == 0:
            return f'{type(self).__name__}([])'
        lens = [len(str(attr)) for attr in attrs]
        align_len = min(max(lens), 20)
        fmt=f'{{:<{align_len}}}'
        contents = [attr._pretty_repr(fmt=fmt) if hasattr(attr, '_pretty_repr') else repr(attr) for attr in attrs]
        contents_str = ',\n  '.join(contents)
        return f'{type(self).__name__}([\n  {contents_str}\n])'


class MetaBehaviorHaver(type):
    '''metaclass for BehaviorHaver. adds cls_behavior_attrs to the class namespace;
    use cls_behavior_attrs.register(new_attr) to register a new attr as a behavior attr.

    Note that __prepare__ gets called before class definition begins, and fills class namespace,
    then __init__ gets called after class definition ends. For example:
        class MySubclass(metaclass=MetaBehaviorHaver):
            # <-- __prepare__ runs here. Roughly: locals().update(__prepare__())
            pass  # or, define other functions, variables, etc, as desired.
        # <-- __init__ runs here. Roughly: MetaBehaviorHaver.__init__(MySubclass)
    '''
    # In case anyone is looking through the codebase to understand cls_behavior_attrs.register
    # this class is the one that defines it, though it doesn't use the usual 'def' notation,
    # so, here are some strings that will pop up for searches like that, instead.
    #   def cls_behavior_attrs(...)            # see __prepare__
    #   def cls_behavior_attrs.register(...)   # see __prepare__, and ClsBehaviorAttrs

    @classmethod
    def __prepare__(_metacls, _name, bases, **kw_super):
        '''return a dict with register_behavior_attr in it.'''
        super_result = super().__prepare__(_name, bases, **kw_super)
        cls_behavior_attrs = ClsBehaviorAttrs()
        for base in bases:
            base_attrs = getattr(base, 'cls_behavior_attrs', [])
            for attr in base_attrs:
                # check for matches!
                for i, cattr in enumerate(cls_behavior_attrs):
                    # check for same str but different object, i.e. registered same str in 2 different subclasses.
                    #   if this occurs we need to be sure to keep the one from the earliest class in the mro().
                    if attr == cattr:
                        if (attr is not cattr) and issubclass(base, cattr.cls_where_defined):
                            cls_behavior_attrs[i] = attr  # attr from earlier in mro() than cattr; replace cattr with attr.
                        break  # (MetaBehaviorHaver logic guarantees there's at most 1 match.)
                else:  # didn't find a match; add attr!
                    cls_behavior_attrs.append(attr)
        return dict(**super_result, cls_behavior_attrs=cls_behavior_attrs)

    def __init__(cls, *args, **kw):
        '''initialize cls (with metaclass=MetaBehaviorHaver)'''
        cls.cls_behavior_attrs.cls_associated_with = cls
        for attr in cls.cls_behavior_attrs:
            if getattr(attr, 'cls_where_defined', None) is None:
                attr.cls_where_defined = cls
            if getattr(attr, 'doc', None) is None:
                if hasattr(cls, attr) and isinstance(getattr(cls, attr), property):
                    attr.doc = getattr(cls, attr).__doc__
        super().__init__(*args, **kw)


class BehaviorHaver(metaclass=MetaBehaviorHaver):
    '''object which has a behavior.'''
    behavior = property(lambda self: self.get_behavior(),
            doc='''dict of {attr: self.attr} for attr in self.behavior_attrs. Note dims are separate;
            dims go in behavior.dims. E.g. Behavior({'units':'si',...}, dims={'snap':0,...}).''')

    def get_behavior(self, keys=None):
        '''return value of self.behavior.
        keys: None or iterable
            if provided, only include these attrs.
            from nondim_behavior_attrs, or dims.
        '''
        attrs = self.nondim_behavior_attrs
        dims = getattr(self, 'dims', [])
        if keys is None:
            use_attrs = attrs
            use_dims = dims
        else:
            use_attrs = [key for key in keys if key in attrs]
            use_dims = [key for key in keys if key in dims]
        behavior_dict = {k: getattr(self, k) for k in use_attrs}
        self_dim_values = getattr(self, 'dim_values', None)
        dims_dict = None if self_dim_values is None else self_dim_values(use_dims)
        return Behavior(behavior_dict, dims_dict)

    @property
    def behavior_attrs(self):
        '''list of attrs in self which control behavior of self.
        Here, returns self.cls_behavior_attrs.

        Subclasses could override if any behavior attrs are not known at the class-level,
            e.g. if MySubclass's list of behavior attrs varies between instances of MySubclass.
        '''
        return self.cls_behavior_attrs

    @property
    def nondim_behavior_attrs(self):
        '''list of attrs in self which control behavior of self, but which are NOT in self.dimensions.'''
        return [attr for attr in self.behavior_attrs if not getattr(attr, 'is_dimension', False)]

    # [TODO] help methods, e.g.:
    # def help_behavior_attrs(self, search) checking all behavior_attrs
    # def help_cls_behavior_attrs(cls, search) as a classmethod checking only cls_behavior_attrs



"""
File Purpose: fft calculations
"""

from ..quantity_loader import QuantityLoader
from ...errors import DimensionError, DimensionKeyError
from ...tools import (
    simple_property, alias, alias_child,
    format_docstring,
    UNSET,
    xarray_fftN, xarray_ifftN, xarray_lowpass,
    FFTDimname, FFTSlices,
)
from ...defaults import DEFAULTS

FFT_VARS = 'xyzt'  # one-letter string coordinate name options for fft.
_MAX_N_FFT_VARS = len(FFT_VARS)   # max number of FFT_VARS in a single call

class FFTLoader(QuantityLoader):
    '''fft calculations, e.g. fft2, fft. 
    self.fft_dims controls which dims to take the fft over.
    '''
    # # # HELPER METHODS # # #
    @format_docstring(xarray_fftN_docs=xarray_fftN.__doc__, sub_ntab=2)
    def fftN(self, array, dim=UNSET, ds=None, *, slices=UNSET, **kw_xarray_fftN):
        '''xarray_fftN with defaults for dim & slices determined by self.fft_dims, self.fft_slices.
        kwargs are passed to xarray_fftN. For convenience, docs for xarray_fftN are copied below.

        xarray_fftN docs
        ----------------
            {xarray_fftN_docs}
        '''
        if dim is UNSET:
            dim = self.fft_dims_for(array)
        if slices is UNSET:
            slices = self.fft_slices
        return xarray_fftN(array, dim, ds=ds, slices=slices, **kw_xarray_fftN)

    fft = alias('fftN')  # alias to fftN, for improved discoverability.

    @format_docstring(xarray_ifftN_docs=xarray_ifftN.__doc__, sub_ntab=2)
    def ifftN(self, array, dim=UNSET, df=None, *, x0=0, ds=None, **kw_xarray_ifftN):
        '''xarray_ifftN with defaults for dim determined by self.fft_dims.
        kwargs are passed to xarray_ifftN. For convenience, docs for xarray_ifftN are copied below.

        xarray_ifftN docs
        -----------------
            {xarray_ifftN_docs}
        '''
        if dim is UNSET:
            dim = self.ifft_dims_for(array)
        return xarray_ifftN(array, dim, df=df, x0=x0, ds=ds, **kw_xarray_ifftN)

    ifft = alias('ifftN')  # alias to ifftN, for improved discoverability.

    @format_docstring(xarray_lowpass_docs=xarray_lowpass.__doc__, sub_ntab=2)
    def lowpass(self, array, dim=UNSET, keep=UNSET, *, keep_r=UNSET, **kw_xarray_lowpass):
        '''xarray_lowpass with defaults for dim & keep determined by self.fft_dims, self.lowpass_keep.
        kwargs are passed to xarray_lowpass. For convenience, docs for xarray_lowpass are copied below.

        xarray_lowpass docs
        -------------------
            {xarray_lowpass_docs}
        '''
        if dim is UNSET:
            dim = self.fft_dims_for(array)
        if keep is UNSET and keep_r is UNSET:
            keep = self.lowpass_keep
        return xarray_lowpass(array, dim, keep=keep, keep_r=keep_r, **kw_xarray_lowpass)

    # # # BEHAVIOR ATTRS # # #
    cls_behavior_attrs.register('fft_dims', getdefault=lambda ql: getattr(ql, 'maindims', []))
    
    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['fft_keep', 'fft_half', 'fft_step'] + any values from super().
        '''
        return ['fft_keep', 'fft_half', 'fft_step'] + super()._extra_kw_for_quantity_loader_call

    @property
    def fft_dims(self):
        '''the dims over which to possibly apply fft (FFTLoader methods).
        will only apply fft along these dims for an array if they actually appear in the array.
        None --> use self.maindims. (this is the default.)
        See also: self.fft_dims_for(array).
        '''
        if getattr(self, '_fft_dims', None) is None:
            return getattr(self, 'maindims', [])
        else:
            return self._fft_dims
    @fft_dims.setter
    def fft_dims(self, value):
        self._fft_dims = value

    def fft_dims_for(self, array):
        '''return the dims over which to apply fft for this array.
        This is the intersection of self.fft_dims and array.dims.
        '''
        return tuple(dim for dim in self.fft_dims if dim in array.dims)

    def ifft_dims_for(self, array):
        '''return dims over which to apply ifft for this array.
        This is the self.fft_dims which correspond to dims in array.dims,
            though not via exact match.
            E.g. 'freq_x' or 'k_x' in array.dims, if 'x' in fft_dims.
        '''
        names = []
        for s in self.fft_dims:
            try:
                names.append(FFTDimname.implied_from(s, array.dims, post_fft=True))
            except DimensionKeyError:
                continue
        return tuple(names)

    _fft_slices_cls = FFTSlices  # cls used for self.fft_slices.

    cls_behavior_attrs.register('fft_slices', default=FFTSlices())
    fft_slices = simple_property('_fft_slices', setdefaultvia='_new_fft_slices',
            doc='''the dict of indexers to apply to all fft results from self, by default.
            keys can be a pre-fft or post-fft dimension name,
                e.g. 'x' or 'freq_x' both lead to slicing of the result's 'freq_x' dimension.
                (note if rad=True it would be 'k_x' in the result, and 'x' would apply but not 'freq_x'.)
                all other keys (not a pre-fft or post-fft dimension name) are ignored.
            values can be slice, int, iterable, or non-integer value between -1 and 1.
                fractional values are interpreted as a fraction of the length of the corresponding dimension,
                as per interprets_fractional_indexing. Negative fractions refer to distance from the end.
                e.g., dict(x=slice(-0.3, None, 0.01), y=0.8), where x and y correspond to length 1000,
                    would be equivalent to dict(x=slice(-300, None, 10), y=800).
            Can also have special keys (which apply to all fft dims without a specifically-related key):
                keep: fraction of each dimension to keep,
                    e.g. keep=0.4 with length=1000 would result in slice(300, 700),
                    since that keeps 400 out of 1000 points, and is centered at 1000/2.
                half: dimension(s) along which to keep only the right half of the result.
                step: slice step. Can also be fractional to use fraction of dimension length.

                For more help on special keys, see help(self._fft_slices_cls) or help(FFTSlices)''')
    @fft_slices.setter
    def fft_slices(self, value):
        '''set self.fft slices. If value is not an instance of FFTSlices, convert it to one.'''
        if not isinstance(value, self._fft_slices_cls):
            if not isinstance(value, dict):
                raise TypeError(f'fft_slices must be a dict, but got type(value)={type(value)}')
            value = self._fft_slices_cls(**value)
        self._fft_slices = value

    def _new_fft_slices(self):
        '''Called when self.fft_slices accessed but not yet set; create & return new FFTSlices.'''
        return self._fft_slices_cls()

    fft_keep = alias_child('fft_slices', 'keep')  # alias to fft_slices.keep
    fft_half = alias_child('fft_slices', 'half')  # alias to fft_slices.half
    fft_step = alias_child('fft_slices', 'step')  # alias to fft_slices.step

    cls_behavior_attrs.register('lowpass_keep', default=DEFAULTS.LOWPASS_KEEP)
    lowpass_keep = simple_property('_lowpass_keep', setdefault=lambda: DEFAULTS.LOWPASS_KEEP,
        doc='''the default value for "keep" in self.lowpass. 0 < lowpass_keep <= 1.
            Or, can be a dict of {dim: keep} pairs, to use different keep for different dims.
            To use keep_r instead of keep, call lowpass directly and enter keep_r there.''')


    # # # LOADABLE QUANTITIES # # #
    @known_pattern(r'(rad)?fft(\d*)_(.+)', deps=[2])  # 'radfft_{var}', 'fft_{var}', 'radfft{N}_{var}', or 'fft{N}_{var}'
    def get_fft(self, var, *, _match=None):
        '''N-dimensional fft. fft(var). Applied along all fft_dims in array.
        self.get('[rad]fft[N]_[var]'), where [rad]='rad' or '', and [N]=any integer or ''
            E.g. 'fft_var', 'radfft_var', 'fft1_var', 'fft2_var', 'radfft1_var', 'radfft2_var'.
        'rad' --> multiply result's frequency coordinates by 2 * pi.
                (array values will be the same either way, but coordinates will be different.)
        N provided --> fft must be along exactly this many dimensions, else crash with DimensionError.
                E.g. N=2 means self(array) must have exactly 2 dims which are also in self.fft_dims.
                Feel free to separately self.fft_dims, or enter it as a kwarg via self(..., fft_dims=...).
        '''
        radstr, Nstr, here = _match.groups()
        array = self(here)
        fft_dims = self.fft_dims_for(array)
        if len(Nstr) > 0:
            N = int(Nstr)
            if len(fft_dims) != N:
                errmsg = f"'{radstr}fft{N}_...' requires exactly {N} dims, but got {len(fft_dims)} dims: {fft_dims}"
                raise DimensionError(errmsg)
        rad = (radstr == 'rad')
        # [TODO] for efficiency, put 'ds' in deps and pass 'ds' values to xarray_fftN,
        #    instead of generating coord array, then making xarray_fftN determine ds from coord array.
        #    (bookkeeping might be a bit tricky since fft_dims aren't guaranteed to be only spatial.)
        return self.fftN(array, fft_dims, rad=rad)

    # 'radfft{dims}_{var}' or 'fft{dims}_{var}'. {dims} can be any dims from FFT_VARS, e.g. 'xy', 'zyt', 'xt', 'xyzt'
    @known_pattern(fr'(rad)?fft([{FFT_VARS}]{{1,{_MAX_N_FFT_VARS}}})_(.+)', deps=[2])
    @format_docstring(FFT_VARS=FFT_VARS)
    def get_fft_with_dims(self, var, *, _match=None):
        '''N-dimensional fft. fft(var). Applied along the indicated dims.
        self.get('([rad]fft[dims]_[var]'), where [rad]='rad' or '', and [dims]=any combination of {FFT_VARS}.
            E.g. 'fftx_var', 'fftt_var', 'fftyz_var', 'radfftxy_var', 'fftzyt_var', 'radfftyzt_var'.
        'rad' --> multiply result's frequency coordinates by 2 * pi.
                (array values will be the same either way, but coordinates will be different.)
        '''
        radstr, fft_dims, here = _match.groups()
        array = self(here)
        fft_dims_list = list(fft_dims)
        dims_match = self.fft_dims_for(array)
        if not all(dim in dims_match for dim in fft_dims):
            errmsg = (f"'{radstr}fft{fft_dims}_...' requires all dims ({fft_dims_list}) to be in self.fft_dims_for(array),"
                      f" but some are missing!"
                      f"\nGot self.fft_dims_for(array)={dims_match}. Note: self.fft_dims={self.fft_dims}, array.dims={array.dims}")
            raise DimensionError(errmsg)
        rad = (radstr == 'rad')
        return self.fftN(array, fft_dims, rad=rad)

    @known_pattern(r'ifft(\d*)_(.+)', deps=[1])  # 'ifft_{var}', 'ifft{N}_{var}'
    def get_ifft(self, var, *, _match=None):
        '''N-dimensional ifft. ifft(var). Applied along all fft_dims in array.
        self.get('ifft[N]_[var]'), where [N]=any integer or ''
            E.g. 'ifft_var', 'ifft1_var', 'ifft2_var'.
        N provided --> ifft must be along exactly this many dimensions, else crash with DimensionError.
                E.g. N=2 means self(array) must have exactly 2 dims which are also in self.fft_dims.
                Feel free to separately self.fft_dims, or enter it as a kwarg via self(..., fft_dims=...).
        '''
        Nstr, here = _match.groups()
        array = self(here)
        dims = self.ifft_dims_for(array)
        if len(Nstr) > 0:
            N = int(Nstr)
            if len(dims) != N:
                errmsg = f"'ifft{N}_...' requires exactly {N} dims, but got {len(fft_dims)} dims: {fft_dims}"
                raise DimensionError(errmsg)
        return self.ifftN(array, dims)

    @known_pattern(r'lowpass_(.+)', deps=[0])  # 'lowpass_{var}'
    @format_docstring(default_lowpass_keep=DEFAULTS.LOWPASS_KEEP)
    def get_lowpass(self, var, *, _match=None):
        '''lowpass filter across self.fft_dims; keep low frequencies, zero high frequencies.
        ifft(fft(self(var) * filter), where filter = 1 for low frequencies, 0 for high frequencies.

        fraction of each fft'd dimension to keep is determined by self.lowpass_keep.
            Default is DEFAULTS.LOWPASS_KEEP (default: {default_lowpass_keep}).
        '''
        here, = _match.groups()
        array = self(here)
        return self.lowpass(array)



"""
Package Purpose: dimensions affecting plasma quantity results
E.g. "first fluid", "vector component axis", "second fluid"
"""
# behaviors
from .behavior import (
    Behavior, BehaviorQuantity,
    BehaviorAttr, ClsBehaviorAttrs, MetaBehaviorHaver, BehaviorHaver,
)
# the DimensionHavers
from .components import Component, ComponentList, ComponentHaver
from .fluids import (
    Fluid, FluidList,
    UniqueFluid, SINGLE_FLUID,
    FluidSpecialValueSpecifier,
    CHARGED, ELECTRON, ELECTRONS, ION, IONS, NEUTRAL, NEUTRALS,
    FluidDimension, FluidHaver,
    jFluidDimension, jFluidHaver,
    FluidsHaver,
)
from .main_dimensions import MainDimensionsHaver
from .snaps import (
    Snap, SnapList, SnapDimension, SnapHaver,
    UniqueSnap, INPUT_SNAP, MISSING_SNAP,
    ProxySnap, ProxySnapList,
    ParamsSnap, ParamsSnapList,
    SELECT_CLOSEST, SELECT_BEFORE, SELECT_AFTER,
    SELECT_ALL_BEFORE, SELECT_ALL_AFTER, SELECT_BETWEEN,
)
from .subsampling import (
    SubsamplingInfo, Subsamplable,
    SubsamplingInfoPathManager, SubsamplingResultPathManager,
    SubsamplingApplier, SubsamplingSlice,
)

# parent class & other tools, for the other dimension havers.
# Probably won't use these directly, outside of this subpackage.
from .dimension_tools import (
    DimensionValue, SliceableList,
    DimensionValueList, string_int_lookup,
    DimensionSpecialValueSpecifier, DimensionSingleValueSpecifier,
    DimensionFractionalIndexer,
    SpecialDimensionValue, UniqueDimensionValue,
    Dimension,
    DimensionHaver,
)

# miscellaneous other stuff
from .components import XYZ, YZ_FROM_X, XHAT, YHAT, ZHAT



"""
File Purpose: binding xarray-related methods to classes defined elsewhere in tools.
(No other xarray_tools file should import this one.)
(This is to avoid circular dependencies.
    E.g. DictOfSimilar is defined in iterables.py,
    and should have the to_ds() method, which relies on some xarray_tools,
    but iterables.py is imported into various xarray_tools files already.)
"""
import xarray as xr

from .xarray_coords import xarray_str_coords
from ..docs_tools import format_docstring
from ..iterables import DictOfSimilar
from ..oop_tools import Binding
from ...errors import InputError, InputConflictError

binding = Binding(locals())


''' --------------------- DictOfSimilar.to_da and .to_ds --------------------- '''

with binding.to(DictOfSimilar):
    # # # CONVERT DictOfSimilar TO XARRAY OBJECT # # #
    @binding
    def to_ds(self, da_dim=None, *, var=None, drop_coords=False, str_coords=False, item=False,
              errors_ok=False):
        '''return Dataset based on values in self.
        The default behavior of this method is equivalent to:
            return xr.Dataset(self)

        The main reason to use to_ds is to do common processing steps on the values first.
        No processing steps will be attempted by default.
        Providing inputs can enable various steps, in this order:
            (1) replace values with value[var]
            (2) convert values to DataArray via value.to_dataarray(da_dim)
            (3) drop some coords from each value
            (4) convert values' coords to strings via xarray_str_coords
            (5) convert values to scalars via value.item()
        Input options described below.

        var: None, str, or iterable of str.
            if provided, will get value[var] for each value in self.
        da_dim: None or str
            if provided, will call value.to_dataarray(da_dim) for each value in self.
        drop_coords: bool, str, or list of strs
            whether to call value.drop_vars(...) for each value in self.
            True --> value.drop_vars(value.coods)
            str or list of strs --> value.drop_vars(drop_coords)
        str_coords: bool, str, or list of strs
            whether to call xarray_str_coords on each value in self.
            True --> converts all coords' values to strs.
            str or list of strs --> xarray_str_coords(value, str_coords, promote=True)
        item: bool
            whether to call value.item() for each value in self.
        errors_ok: bool
            whether to ignore simple errors during processing.
            if True:
                var non-None but value not dict or Dataset --> skip this step.
                da_dim non-None but values are not Datasets --> skip this step.
                drop_coords --> value.drop_vars(drop_coords, errors='ignore').
                str_coords --> xarray_str_coords(value, str_coords, missing_ok=True).
                item but value.item() fails with ValueError or AttributeError --> skip this step.

        Example:
            mc = pc.MultiCalculator(...)
            dos = mc('mean_n', snap=7)   # depends on 'fluid' dim, but also has 'snap' and 't' coords.
            n_vals = dos.to_ds(str_coords='fluid', drop_coords=('snap', 't'))
            # str_coords='fluid' because 'fluid' might be different objects for each calculator
            # drop_coords=('snap', 't') because 'snap' and 't' info won't necessarily agree across all calculators.
        '''
        result = {k: v for k, v in self.items()}
        if var is not None:
            if isinstance(var, str):
                var = [var]
            for k, v in result.items():
                if not isinstance(v, dict) and not isinstance(v, xr.Dataset):
                    if errors_ok:
                        continue
                    else:
                        errmsg = (f'provided var={var!r}, but value not a dict or Dataset, and errors_ok=False.'
                                  f' Got value of type={type(v)}, at key={k!r}.')
                        raise InputError(errmsg)
                result[k] = v[var]
        if da_dim is not None:
            for k, v in result.items():
                if not isinstance(v, xr.Dataset):
                    if errors_ok:
                        continue
                    else:
                        errmsg = (f'provided da_dim={da_dim!r}, but value not a Dataset, and errors_ok=False.'
                                  f' Got value of type={type(v)}, at key={k!r}.')
                        raise InputError(errmsg)
                result[k] = v.to_dataarray(da_dim)
        if drop_coords:
            for k, v in result.items():
                to_drop = list(v.coords) if drop_coords == True else drop_coords
                try:
                    result[k] = v.drop_vars(to_drop, errors='ignore' if errors_ok else 'raise')
                except Exception as e:
                    if not errors_ok:
                        errmsg = f'provided drop_coords={drop_coords!r}, but crashed during drop_vars, and errors_ok=False.'
                    else:
                        errmsg = f"crash in drop_vars({drop_coords!r}, errors='ignore') -- not sure why."
                    raise Exception(errmsg) from e
        if str_coords:
            for k, v in result.items():
                try:
                    result[k] = xarray_str_coords(v, str_coords, promote=True, missing_ok=errors_ok)
                except Exception as e:
                    if not errors_ok:
                        errmsg = (f'provided str_coords={str_coords!r}, but crashed during xarray_str_coords, '
                                  f'and errors_ok=False.')
                    else:
                        errmsg = f"crash in xarray_str_coords(arr, {str_coords!r}, missing_ok=True) -- not sure why."
                    raise Exception(errmsg) from e
        if item:
            for k, v in result.items():
                try:
                    result[k] = v.item()
                except (ValueError, AttributeError) as e:
                    if not errors_ok:
                        errmsg = (f'provided item=True, but crashed during value.item(), and errors_ok=False.'
                                  f' Got value of type={type(v)}, at key={k!r}.')
                        raise InputError(errmsg) from e
        return xr.Dataset(result)

    @binding
    @format_docstring(to_ds_doc=to_ds.__doc__)
    def to_da(self, dim='variable', *, ds=False,
              da_dim=None, var=None, drop_coords=False, str_coords=False, item=False,
              errors_ok=False, **kw_to_ds):
        '''return DataArray based on values in self.
        The default behavior of this method is equivalent to:
            return xr.Dataset(self).to_dataarray(dim=dim)

        The other inputs enable to do some processing steps on self.values first.
        If passing any inputs other than `dim` and `ds`, self.to_da(dim, **kw) is equivalent to:
            return self.to_ds(**kw).to_dataarray(dim=dim)

        dim: str
            name of new dimension, across which to stack the data_vars from the dataset.
        ds: bool
            whether to convert result back into a Dataset.
            shorthand for the pattern:
                self.to_ds(dim, da_dim='dadim', **kw).to_dataset('dadim')
            Useful e.g. if self.values are datasets already,
                and the goal is to end up with a Dataset just like each value in self,
                but with the result also having the new `dim` dimension (with size==len(self))
            if True, cannot also provide `da_dim`, and internally uses da_dim='__da_dim_internal__'

        Docs for to_ds() copied below, for convenience:
        -----------------------------------------------
        {to_ds_doc}
        '''
        if ds:
            if da_dim is not None:
                raise InputConflictError('cannot provide both ds=True and da_dim!=None.')
            else:
                da_dim = '__da_dim_internal__'
        as_ds = self.to_ds(da_dim=da_dim, var=var, drop_coords=drop_coords,
                        str_coords=str_coords, item=item, errors_ok=errors_ok, **kw_to_ds)
        result = as_ds.to_dataarray(dim=dim)
        if ds:
            result = result.to_dataset(da_dim)
        return result

    @binding
    def to_xr(self, dim='variable', da_dim=None, *,
              var=None, drop_coords=False, str_coords=False, item=False,
              errors_ok=False, **kw_to_ds):
        '''return xarray.DataArray or Dataset, stacking values in self along the new dimension.
        This is shorthand for self.to_da() or self.to_ds()
            (depending on `da_dim` and whether self.values are already xarray.Dataset objects).

        dim: str
            the new dimension, along which to stack values from self.
        da_dim: None or str
            if provided, will call value.to_dataarray(da_dim) for each value in self.

        If da_dim provided or if there are no Dataset values in self,
            result will be a DataArray, as per self.to_da(dim, ds=False, da_dim=da_dim, **kw).
        Otherwise, if all self.values are Datasets,
            result will be a Dataset, as per self.to_da(dim, ds=True, da_dim=None).
        Otherwise (only some values are Datasets, and da_dim not provided),
            crash with InputError.

        The remaining kwargs go to self.to_ds() (see help(self.to_ds) for details):
            var: None, str, or iterable of str.
                if provided, will get value[var] for each value in self.
            drop_coords: bool, str, or list of strs
                whether to call value.drop_vars(...) for each value in self.
            str_coords: bool, str, or list of strs
                whether to call xarray_str_coords on each value in self.
            item: bool
                whether to call value.item() for each value in self.
            errors_ok: bool
                whether to ignore simple errors during processing.
        '''
        kw_to_ds.update(var=var, drop_coords=drop_coords, str_coords=str_coords, item=item, errors_ok=errors_ok)
        if (da_dim is not None) or not any(isinstance(v, xr.Dataset) for v in self.values()):
            return self.to_da(dim=dim, ds=False, da_dim=da_dim, **kw_to_ds)
        elif all(isinstance(v, xr.Dataset) for v in self.values()):
            return self.to_da(dim=dim, ds=True, da_dim=None, **kw_to_ds)
        else:
            errmsg = ('to_xr cannot decide whether to return DataArray or Dataset, '
                      'when some values are Datasets and some are not, and da_dim=None.\n'
                      'Consider using to_da or to_ds directly instead.')
            raise InputConflictError(errmsg)



"""
File Purpose: xarray stats, math, and any other aggregation functions.
(aggregation: reducing array along one or more dims)
"""
import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from .xarray_dimensions import (
    xarray_ensure_dims, _paramdocs_ensure_dims,
)
from ..docs_tools import format_docstring
from ...errors import InputConflictError
from ...defaults import DEFAULTS


''' --------------------- Docstrings --------------------- '''

_paramdocs_aggregator = {
    'dim': '''None, str, or iterable of strs
        apply operation along these dimensions''',
    'keep': '''None, str, or iterable of strs
        apply operation along all except for these dimensions.
        (can provide keep or dim, but not both.)''',
    'promote_dims_if_needed': _paramdocs_ensure_dims['promote_dims_if_needed'],
    'missing_dims': f'''{_paramdocs_ensure_dims['missing_dims']}
        if not 'raise', any missing dims will be skipped.''',
}


''' --------------------- Aggregators setup --------------------- '''

@format_docstring(**_paramdocs_aggregator)
def _xarray_aggregator_setup(array, dim=None, keep=None, promote_dims_if_needed=True, missing_dims='raise'):
    '''returns (array, dim, keep), adjusted appropriately.
    adjustments include:
        - result dim & keep will be None or list of str. (replace str input s with [s])
        - if keep provided, infer dim from array.
        - can provide dim or keep but not both.
        - (if dim or keep provided) ensure all dim in array.dims.
            promote relevant scalar coords if promote_dims_if_needed.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}
    '''
    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
    if dim is not None and keep is not None:
        raise InputConflictError('cannot provide both "dim" and "keep".')
    if keep is not None:
        keep = [keep] if isinstance(keep, str) else list(keep)
        dim = set(array.dims) - set(keep)
    if dim is not None:
        dim = [dim] if isinstance(dim, str) else list(dim)
        array, dim = xarray_ensure_dims(array, dim, promote_dims_if_needed=promote_dims_if_needed,
                                        missing_dims=missing_dims, return_existing_dims=True)
    return array, dim, keep

@pcAccessor.register('aggregate', aliases=['agg'])
@format_docstring(**_paramdocs_aggregator)
def xarray_aggregate(array, f, dim=None, *, keep=None,
                     promote_dims_if_needed=True, missing_dims='raise', **kw_agg_f):
    '''returns array aggregated along dim using f.

    f: callable or str
        function to aggregate along dim. E.g. xarray.DataArray.sum, or 'mean'.
        Call signature should be like: f(array, dim, **kw).
        str --> will call getattr(array, f)(dim, **kw).
    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to f.
    '''
    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
    array, dim, keep = _xarray_aggregator_setup(array, dim, keep, promote_dims_if_needed, missing_dims)
    if isinstance(f, str):
        g = getattr(array, f)
        return g(dim, **kw_agg_f)
    else:
        return f(array, dim, **kw_agg_f)


''' --------------------- Math (except stats) --------------------- '''

@pcAccessor.register('sum')
@format_docstring(**_paramdocs_aggregator)
def xarray_sum(array, dim=None, *, keep=None,
               promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''Like xarray sum but first promotes dims if needed, and accepts 'keep' option.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.sum.
    '''
    return xarray_aggregate(array, 'sum', dim=dim, keep=keep,
                            promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims, **kw)


''' --------------------- Stats --------------------- '''

@pcAccessor.register('stats')
@format_docstring(**_paramdocs_aggregator)
def xarray_stats(array, dim=None, stats=None, *, keep=None,
                 promote_dims_if_needed=True, missing_dims='raise',
                 to_da=None, **kw):
    '''returns Dataset of stats for array: min, mean, median, max, std, rms.

    array: xarray.DataArray or xarray.Dataset
        the array (or dataset) from which to compute stats.
        if dataset, compute separate stats for each data var,
            and result will be a dataset with 'stat' dim;
            equivalent to doing xarray_stats for each data_var, with `to_da`='stat'.
    dim: {dim}
    stats: None or iterable of str from ('min', 'mean', 'median', 'max', 'std', 'rms')
        which stats to compute & include in the result.
        None --> equivalent to stats=['min', 'mean', 'median', 'max', 'std', 'rms'].
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}
    to_da: None, bool, or str
        whether to return DataArray instead of Dataset.
        None --> True if `array` is a Dataset, else False.
        False --> result is a Dataset with stats as data vars.
                  (incompatible with input `array` being a Dataset)
        True --> equivalent to to_da='stat'.
        str --> stats reported along this dimension (must not already exist in `array`).
                result type is the same type as input (DataArray or Dataset).

    additional kwargs are passed to array.min, mean, median, max, and std.
    Can provide skipna=True/False to skip NaNs or not; default True for float dtypes.
    '''
    array, dim, keep = _xarray_aggregator_setup(array, dim, keep, promote_dims_if_needed, missing_dims)
    stats = ['min', 'mean', 'median', 'max', 'std', 'rms'] if stats is None else list(stats)
    results = dict()
    if 'min' in stats: results['min'] = array.min(dim, **kw)
    if 'mean' in stats: results['mean'] = array.mean(dim, **kw)
    if 'median' in stats: results['median'] = array.median(dim, **kw)
    if 'max' in stats: results['max'] = array.max(dim, **kw)
    if 'std' in stats: results['std'] = array.std(dim, **kw)
    if 'rms' in stats: results['rms'] = (array**2).mean(dim, **kw)**0.5
    results = {k: results[k] for k in stats}  # <-- style: put in same order as stats.
    # convert to appropriate format for output
    if to_da is None:
        to_da = isinstance(array, xr.Dataset)
    if to_da == False:
        if isinstance(array, xr.Dataset):
            raise InputConflictError('cannot use to_da=False when input is a Dataset.')
        attrs = array.attrs.copy() if xr.get_options()['keep_attrs'] else None
        return xr.Dataset(results, attrs=attrs)
    # <-- if we reached this point, it means we will concatenate. to_da is True (or str).
    if not isinstance(to_da, str):
        to_da = 'stat'
    if to_da in array.coords:
        raise InputConflictError(f'to_da={to_da!r} already in array.coords={set(array.coords)}')
    results = {k: results[k].assign_coords({to_da: k}) for k in results}
    return xr.concat(list(results.values()), dim=to_da)

@pcAccessor.register('min')
@format_docstring(**_paramdocs_aggregator)
def xarray_min(array, dim=None, *, keep=None,
               promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''Like xarray min but first promotes dims if needed, and accepts 'keep' option.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.min.
    '''
    return xarray_aggregate(array, 'min', dim=dim, keep=keep,
                            promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims, **kw)

@pcAccessor.register('mean')
@format_docstring(**_paramdocs_aggregator)
def xarray_mean(array, dim=None, *, keep=None,
                promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''Like xarray mean but first promotes dims if needed, and accepts 'keep' option.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.mean.
    '''
    return xarray_aggregate(array, 'mean', dim=dim, keep=keep,
                            promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims, **kw)

@pcAccessor.register('median')
@format_docstring(**_paramdocs_aggregator)
def xarray_median(array, dim=None, *, keep=None,
                  promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''Like xarray median but first promotes dims if needed, and accepts 'keep' option.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.median.
    '''
    return xarray_aggregate(array, 'median', dim=dim, keep=keep,
                            promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims, **kw)

@pcAccessor.register('max')
@format_docstring(**_paramdocs_aggregator)
def xarray_max(array, dim=None, *, keep=None,
               promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''Like xarray max but first promotes dims if needed, and accepts 'keep' option.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.max.
    '''
    return xarray_aggregate(array, 'max', dim=dim, keep=keep,
                            promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims, **kw)

@pcAccessor.register('std')
@format_docstring(**_paramdocs_aggregator)
def xarray_std(array, dim=None, *, keep=None,
               promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''Like xarray std but first promotes dims if needed, and accepts 'keep' option.

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs are passed to array.std.
    '''
    return xarray_aggregate(array, 'std', dim=dim, keep=keep,
                            promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims, **kw)

@pcAccessor.register('rms')
@format_docstring(**_paramdocs_aggregator)
def xarray_rms(array, dim=None, *, keep=None,
               promote_dims_if_needed=True, missing_dims='raise', **kw):
    '''(array**2).mean(dim, **kw)**0.5

    dim: {dim}
    keep: {keep}
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}
    '''
    mean2 = xarray_mean(array**2, dim=dim, keep=keep,
                        promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims, **kw)
    return mean2**0.5


''' --------------------- Non-aggregators --------------------- '''

def xarray_minimum(array0, *arrays):
    '''minimum across these arrays. Repeatedly apply np.minimum.
    if only input one array, return it, unchanged.
    '''
    if len(arrays) == 0:
        return array0
    result = array0
    for array in arrays:
        result = np.minimum(result, array)
    return result

@pcAccessor.register('minimum', totype='dataset')
def xarray_minimum_of_datavars(ds, *varnames):
    '''minimum across these datavars from ds. Repeatedly apply np.minimum.
    0 varnames --> minimum across all datavars.
    1 varname --> return ds[varname]
    2+ varnames --> minimum across these varnames.
                    equivalent: ds.to_dataarray()[varnames].min(dim='variable')
    '''
    if len(varnames) == 0:
        varnames = list(ds.data_vars)
    vals = [ds[varname] for varname in varnames]
    return xarray_minimum(*vals)

def xarray_maximum(array0, *arrays):
    '''maximum across these arrays. Repeatedly apply np.maximum.
    if only input one array, return it, unchanged.
    '''
    if len(arrays) == 0:
        return array0
    result = array0
    for array in arrays:
        result = np.maximum(result, array)
    return result

@pcAccessor.register('maximum', totype='dataset')
def xarray_maximum_of_datavars(ds, *varnames):
    '''maximum across these datavars from ds. Repeatedly apply np.maximum.
    0 varnames --> maximum across all datavars.
    1 varname --> return ds[varname]
    2+ varnames --> maximum across these varnames.
                    equivalent: ds.to_dataarray()[varnames].max(dim='variable')
    '''
    if len(varnames) == 0:
        varnames = list(ds.data_vars)
    vals = [ds[varname] for varname in varnames]
    return xarray_maximum(*vals)



"""
File Purpose: fluids for multifluid analysis of single-fluid mhd.
"""
import abc

from ..elements import Element, ElementList
from ...defaults import DEFAULTS
from ...dimensions import Fluid, FluidList, FluidSpecialValueSpecifier
from ...errors import InputError, FluidKeyError, FluidValueError
from ...tools import (
    alias, alias_child, simple_property, elementwise_property,
    UNSET, NO_VALUE,
    Binding, identity,
    is_iterable,
    as_roman_numeral,
)
binding = Binding(locals())


''' --------------------- MhdFluid objects --------------------- '''

# # # ELEMENT HANDLER LOGIC # # #

class ElementHandler():
    '''MhdFluid which handles Elements, e.g. might have non-None self.element.
    provides get_element, get_elements, neutral, ion, and saha_list.
    '''
    element_cls = Element  # assert self.get_element() returns an instance of this class, if not None.
    # cls for making new Specie or SpecieList based on element or elements: 
    specie_cls = NotImplemented   # Specie; assigned later in this file.
    specie_list_cls = NotImplemented   # SpecieList; assigned later in this file.

    def get_element(self, default=NO_VALUE):
        '''return self.element if it exists and is not None, else crash with FluidKeyError.
        default: NO_VALUE or any value.
            If provided, when element missing or None, return default instead of crash.
        '''
        element = getattr(self, 'element', None)
        if element is None:
            if default is NO_VALUE:
                raise FluidKeyError(f'get_element() when self.element missing or None. (self={self!r})')
        elif not isinstance(element, self.element_cls):
            raise FluidValueError(f'get_element() expected {self.element_cls.__name__}; got {element!r}')
        return element

    def get_elements(self, default=NO_VALUE):
        '''return list of elements in self.
        The implementation here just returns [self.get_element(default=default)].
        '''
        return [self.get_element(default=default)]

    def neutral(self, **kw_specie):
        '''return neutral Specie of self.get_element().'''
        return self.specie_cls.neutral(self.get_element(), **kw_specie)

    def ion(self, q=1, **kw_specie):
        '''return ionized Specie of self.get_element()
        q: charge in elementary charge units. (default: +1)
        '''
        return self.specie_cls.ion(self.get_element(), q=q, **kw_specie)

    get_I = alias('neutral')
    get_II = alias('ion')

    def saha_list(self, *, istart=0, **kw_specie):
        '''return SpecieList of neutral & once-ionized ions of self.get_element().
        istart: start index for the SpecieList. Index affects conversion to int.
        '''
        neutral = self.neutral(**kw_specie)
        ion = self.ion(**kw_specie)
        return self.specie_list_cls([neutral, ion], istart=istart)


# # # MHDFLUID AND ELEMENTHAVER # # # 

class MhdFluid(ElementHandler, Fluid, abc.ABC):
    '''abstract base class for any fluid in multifluid analysis of single-fluid mhd.
    E.g. Specie or IonMixture.

    Every MhdFluid is an ElementHandler (e.g. has self.get_element(), might return None or crash)
        however not necessarily an ElementHaver (has self.element, probably non-None)
    '''
    # allow ordering of MhdFluids even if they aren't subclasses of each other (e.g. Specie & IonMixture).
    # super() gives NotImplemented if not subclasses of each other;
    # here only gives NotImplemented if not subclasses of MhdFluid.
    def __lt__(self, other):
        if isinstance(other, MhdFluid): return (self.i, self.s) < (other.i, other.s)
        else: return NotImplemented
    def __le__(self, other):
        if isinstance(other, MhdFluid): return (self.i, self.s) <= (other.i, other.s)
        else: return NotImplemented
    def __gt__(self, other):
        if isinstance(other, MhdFluid): return (self.i, self.s) > (other.i, other.s)
        else: return NotImplemented
    def __ge__(self, other):
        if isinstance(other, MhdFluid): return (self.i, self.s) >= (other.i, other.s)
        else: return NotImplemented


class ElementHaver(MhdFluid):
    '''fluid which may "have" an element. E.g. Element or Specie.

    name: None or str
        name of this fluid. None --> name unknown.
    i: None or int
        index of this fluid (within an ElementHaverList)
    element: None, dict, or Element
        the element which this ElementHaver "has".
        dict --> convert to Element via Element.from_dict(element).
    '''
    def __init__(self, name=None, i=None, *, element=None, **kw_super):
        super().__init__(name=name, i=i, **kw_super)
        # (super(MhdFluid) because we want to skip MhdFluid's __init__ which causes crash.)
        if (element is not None) or (not hasattr(self, 'element')):
            if (element is not None) and not isinstance(element, Element):
                element = Element.from_dict(element)
            self.element = element

    def to_dict(self, *, element_to_dict=True):
        '''return dictionary of info about self. Attribute values for keys in self._kw_def.
        e.g. if _kw_def={'name', 'i', 'm'}, then result = {'name': self.name, 'i': self.i, 'm': self.m}

        element_to_dict: bool, default True
            whether to also convert result['element'].to_dict() too, if non-None 'element' in result.
        '''
        result = super().to_dict()
        if element_to_dict and (result.get('element', None) is not None):
            result['element'] = result['element'].to_dict()
        return result


# # # ELEMENT CLASS HOOKUP # # #
# Element is a "virtual subclass" of ElementHaver, instead of a true subclass,
# to keep the elements.py code focused on single-fluid mhd,
# without worrying about multifluid complications.

ElementHaver.register(Element)  # (see abc.ABC for details)
with binding.to(Element):  # assign most of the relevant ElementHandler methods & attributes:
    # cls for making new Element, Specie, or SpecieList objects:
    Element.element_cls = Element
    Element.specie_cls = NotImplemented   # Specie; assigned later in this file.
    Element.specie_list_cls = NotImplemented   # SpecieList; assigned later in this file.
    # methods that are exactly the same for Element and ElementHaver:
    Element.neutral   = ElementHaver.neutral
    Element.ion       = ElementHaver.ion
    Element.get_I     = ElementHaver.get_I
    Element.get_II    = ElementHaver.get_II
    Element.saha_list = ElementHaver.saha_list

    # stuff defined here:
    Element.element = property(identity,
        doc='''alias to self. because self (an Element) is the element corresponding to self.''')

    @binding
    def get_element(self, *args__None, **kw__None):
        '''return self (because self is an Element...).
        (ElementHaver-related logic expects this method to exist.)
        '''
        return self

    @binding
    def get_elements(self, *args__None, **kw__None):
        '''return [self] (because self is an Element...).
        (ElementHaver-related logic expects this method to exist.)
        '''
        return [self]

    # no need to bind to_dict; the existing Element.to_dict (from super()) works just fine.


# # # SPECIE # # #
# [TODO] make GenericSpecie for electron, neutral, or ion,
#   and make ElectronSpecie (for electron) and Specie (for non-electron) subclasses,
#   so that Specie can actually require element to be provided.

class Specie(ElementHaver):
    '''Fluid corresponding to a specie, e.g. H_I, H_II, e, Fe_II.
    (code uses 'specie' as singular of 'species', to clarify singular vs plural.)

    CAUTION: some code assumes (without checking) that Specie objects are immutable.
        Changing a Specie after creating it could lead to unexpected behavior.

    species usually have a name. Every specie has a charge (possibly charge=0)
    non-electron species should also have an associated Element.

    name: UNSET, None, or str
        name of this specie.
        UNSET --> infer from element and q; None if no element.
                inferred name will be element name + roman numeral based on q,
                e.g. H_I for neutral H, H_II for H+.
        None --> cannot convert self to str.
    i: None or int
        the index of this specie (within a SpecieList).
        None --> cannot convert self to int.
    q: number
        charge, in elementary charge units (e.g. -1 for electrons, +1 for H_II)
    m: UNSET, None, or number
        mass, in atomic mass units (e.g. ~1 for H_I or H_II).
        UNSET --> use element.m.
        None --> cannot get self.m
    element: None, dict, or Element
        Element associated with this specie.
        dict --> convert to Element via Element.from_dict(element).
    '''
    _kw_def = {'name', 'i', 'q', 'm', 'element'}

    def __init__(self, name=UNSET, i=None, *, q, m=UNSET, element=None):
        if (element is not None) and not isinstance(element, Element):
            element = Element.from_dict(element)
        if name is UNSET:
            name  = None if element is None else f'{element.name}_{as_roman_numeral(q+1)}'
        if m is UNSET:
            m     = None if element is None else element.m
        super().__init__(name=name, i=i, m=m, q=q, element=element)

    # # # CREATION OPTIONS # # #
    @classmethod
    def electron(cls, name='e', i=None, *, m=UNSET, **kw_init):
        '''create an electron Specie (with q=-1).
        m: if UNSET, use physical value: DEFAULTS.PHYSICAL.CONSTANTS_SI['me amu-1']
        '''
        if m is UNSET: m = DEFAULTS.PHYSICAL.CONSTANTS_SI['me amu-1']
        return cls(name=name, i=i, q=-1, m=m, **kw_init)

    @classmethod
    def from_element(cls, element, i=None, *, q, **kw_init):
        '''create a Specie from this Element (& charge q [elementary charge units]).
        name and m inferred from element and q, unless provided explicitly here.
        '''
        return cls(i=i, q=q, element=element, **kw_init)

    @classmethod
    def neutral(cls, element, i=None, **kw_init):
        '''create a neutral Specie (with q=0) from element.
        element: Element object. Used to infer name and m, by default.
        '''
        return cls.from_element(element, i=i, q=0, **kw_init)

    @classmethod
    def ion(cls, element, i=None, *, q=1, **kw_init):
        '''create an ion Specie from element and q [elementary charge units]
        element: Element object. Used to infer name and m, by default.
        '''
        if q <= 0:
            raise FluidValueError(f'q must be positive for ions, but got q={q}')
        return cls.from_element(element, i=i, q=q, **kw_init)

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [repr(val) for val in [self.name, self.i] if val is not None]
        contents.append(f'q={self.q}')
        if (self.m is not None) and ((self.element is None) or (self.m != self.element.m)):
            contents.append(f'm={self.m:{".1e" if self.m < 1e-3 else ".3f"}}')
        if self.element is not None:
            contents.append(f'element={self.element.name}')
        return f'{type(self).__name__}({", ".join(contents)})'


# # # ION MIXTURE # # #

class IonMixture(MhdFluid):
    '''Fluid corresponding to multiple ions grouped together, e.g. Mg+Al+Si_II

    CAUTION: some code assumes (without checking) that IonMixture objects are immutable.
        Changing an IonMixture after creating it could lead to unexpected behavior.

    IonMixtures usually have a name. Every IonMixture has a charge (probably charge=1).
    Name and charge can be inferred from the species in the mixture.

    species: iterable of Specie or dict objects
        the list of Species in this mixture. length must be at least 1.
        internally stored as SpecieList(species)
        dicts --> convert to SpecieList via SpecieList.from_dicts(species).
    i: None or int
        the index of this mixture (within an IonMixtureList or MhdFluidList)
        None --> cannot convert self to int.
    q: UNSET or number
        charge, in elementary charge units (e.g. +1 for H_II).
        all species must have the same charge else raise NotImplementedError.
        UNSET --> infer q from species.
        number --> assert q > 0, and q == spec.q for spec in species.
    name: UNSET, None, or str
        name of this mixture. None --> unnamed.
        UNSET --> infer from species. E.g. Mg_II, Al_II, Si_II --> Mg+Al+Si_II
    m_mean_mode: str in {'simple', 'density'}
        how to calculate the mass of this mixture.
        'simple' --> m = mean(spec.m for spec in species). This is a constant.
        'density' --> m = density weighted mean of spec.m for spec in species.
                          this might vary across maindims and snaps...
    m_tol: UNSET, None, or number
        crash if any |m_s - mean(m)| is more than m_tol * mean(m).
        (using simple mean for this test, regardless of m_mean_mode)
        (Prevents accidentally creating a mixture of ions with very different masses.)
        UNSET --> use self.DEFAULT_M_TOL (default=0.1)
        None --> allow any masses.
    '''
    _kw_def = {'species', 'i', 'q', 'name', 'm_mean_mode', 'm_tol'}
    _kw_eq = {'species', 'i', 'q', 'name', 'm_mean_mode'}  # (don't check m_tol when testing for equality)

    specie_cls = Specie  # ensure all species are instances of this class.
    specie_list_cls = NotImplemented  # SpecieList; assigned later in this file

    # crash if |m_s - mean(m)| > m_tol * mean(m).
    DEFAULT_M_TOL_SIMPLE = 0.1  # default m_tol when m_mean_mode='simple'
    DEFAULT_M_TOL_DENSITY = 0.3  # default m_tol when m_mean_mode='density'
    # m_tol density default larger because density-weighting is probably more physically accurate,
    #    e.g. if one of the species' is not present it will not contribute at all.
    
    def __init__(self, species, i=None, *, q=UNSET, name=UNSET,
                 m_mean_mode='simple', m_tol=UNSET):
        if len(species) == 0:
            raise InputError(f'{type(self).__name__} requires len(species)>0')
        if isinstance(species[0], dict):
            species = self.specie_list_cls.from_dicts(species)
        else:
            species = self.specie_list_cls(species)
        if self.specie_cls is not None:
            if any(not isinstance(spec, self.specie_cls) for spec in species):
                errmsg = (f'{type(self).__name__} expects all species to be '
                          f'instances of {self.specie_cls.__name__}, but got {species}')
                raise FluidValueError(errmsg)
        self.species = species
        if name is UNSET:
            name = self.infer_name()
        if q is UNSET:
            q = species[0].q
        charges = [spec.q for spec in species]
        if any(charge != q for charge in charges[1:]):
            raise NotImplementedError(f"{type(self).__name__} if species charges differ: {charges}")
        if q is None or q <= 0:
            raise InputError(f'{type(self).__name__} requires q > 0, but got q={q}')
        self.m_mean_mode = m_mean_mode
        if m_tol is UNSET:
            m_tol = self.DEFAULT_M_TOL_SIMPLE if m_mean_mode == 'simple' else self.DEFAULT_M_TOL_DENSITY
        self.m_tol = m_tol
        self._check_m_tol()
        if m_mean_mode == 'simple':
            m = self.m_mean
        else:
            m = None  # mass not constant
        super().__init__(name=name, i=i, q=q, m=m)

    def _check_m_tol(self):
        '''crash if any |m_s - mean(m)| is more than m_tol * mean(m).
        return max(|m_s - mean(m)| / mean(m)).
        '''
        m_tol = self.m_tol
        m_mean = self.m_mean
        largest_diff = 0
        for spec in self.species:
            diff = abs(spec.m - m_mean)
            if (m_tol is not None) and (diff > m_tol * m_mean):
                errmsg = (f'{type(self).__name__} specie mass is too far from mean, for specie {spec}.\n'
                          f'Require |m_s - mean(m)| < m_tol * mean(m), when provided non-None m_tol.\n'
                          f'(m_s={spec.m:.1f}, mean(m)={m_mean:.1f},'
                          f' m_tol={m_tol:.3f}, m_tol*mean(m)= {m_tol*m_mean:.1f})')
                raise FluidValueError(errmsg)
            largest_diff = max(largest_diff, diff / m_mean)
        return largest_diff

    _VALID_M_MEAN_MODES = ('simple', 'density')
    m_mean_mode = simple_property('_m_mean_mode', validate_from='_VALID_M_MEAN_MODES',
            doc='''str telling how to calculate the mass of this mixture.
            'simple' --> m = mean(spec.m for spec in species). This is a constant.
            'density' --> m = density weighted mean of spec.m for spec in species.
                              this might vary across maindims and snaps...''')

    @property
    def m_mean(self):
        '''return mean mass of species in this mixture.
        (Always using a simple mean: sum(s.m for s in species)/len(species))
        '''
        if any(spec.m is None for spec in self.species):
            raise FluidValueError(f'{type(self).__name__}.m_mean needs non-None m for all species.')
        return sum(spec.m for spec in self.species) / len(self.species)

    def infer_name(self):
        '''infer name based on species' names.
        if all species' names have the same suffix (starting at underscore),
            '+'.join(names without suffix) + suffix.
        else, '+'.join(names).
        '''
        prefixes = []
        suffix = None
        names = [spec.name for spec in self.species]
        if all('_' in name for name in names):
            for name in names:
                prefix, _, suffix_here = name.rpartition('_')
                if suffix is None:
                    suffix = suffix_here
                elif suffix != suffix_here:
                    suffix = None
                    break
                prefixes.append(prefix)
        if suffix is None:
            return '+'.join(names)
        else:
            return '+'.join(prefixes) + '_' + suffix

    # temporarily disabled... xarray doesn't like using IonMixture as coord if these are enabled...
    # def __iter__(self):
    #     return iter(self.species)
    # def __len__(self):
    #     return len(self.species)
    # def __getitem__(self, key):
    #     return self.species[key]
    
    # # # DIMENSION VALUE BEHAVIOR # # #
    get = alias_child('species', 'get', doc='''return self.species.get(key)''')

    def lookup_dict(self):
        '''return dict for looking up self within a DimensionValueList, given int, str, or self.
        (used by DimensionValueList.lookup_dict)
        in addition to being able to lookup self,
            also allows lookup of any species within self, by name or object (but not index).
        E.g. if H_II=Specie('H_II', i=3) in self,
            result will include {'H_II': H_II, H_II: H_II}, but will not include {3: H_II}
        '''
        result = super().lookup_dict()
        for spec in self.species:
            result[spec] = spec
            result[str(spec)] = spec
        return result

    def to_dict(self, *, species_to_dict=True):
        '''return dictionary of info about self. Attribute values for keys in self._kw_def.

        species_to_dict: bool, default True
            whether to also convert result['species'].to_dict() too.
        '''
        result = super().to_dict()
        if species_to_dict:
            result['species'] = result['species'].to_dict()  # SpecieList has to_dict() method.
        return result

    # # # ELEMENT HANDLING # # #
    def get_elements(self, default=NO_VALUE):
        '''return list of elements in self, i.e. spec.get_element() for spec in self.'''
        return [spec.get_element(default=default) for spec in self.species]

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [repr(val) for val in [self.name, self.i] if val is not None]
        contents.append(f'q={self.q}')
        if self.m is not None:
            contents.append(f'm={self.m:{".1e" if self.m < 1e-3 else ".3f"}}')
        if self.m_mean_mode != 'simple':
            contents.append(f'm_mean_mode={self.m_mean_mode!r}')
        return f'{type(self).__name__}({", ".join(contents)})'


''' --------------------- MhdFluidList objects --------------------- '''

# # # ELEMENT HANDLER LIST LOGIC # # #

class ElementHandlerList():
    '''list of ElementHandler objects'''
    element_list_cls = ElementList  # class for making new ElementList

    element = elementwise_property('element',
            doc='''[fluid.element for fluid in self].
            To ignore fluids missing elements, consider self.unique_elements() instead.''')
    elements = alias('element')

    def unique_elements(self, *, istart=None):
        '''return ElementList composed of unique elements from ElementHandlers in self.
        istart: None or int. int --> use result[k].with_i(istart + k) for all k.
        '''
        element_lists = [h.get_elements(default=None) for h in self]
        elements = [el for ellist in element_lists for el in ellist if el is not None]
        return self.element_list_cls.unique_from(elements, istart=istart)


# # # MHDFLUIDLIST # # #

class MhdFluidList(ElementHandlerList, FluidList, abc.ABC):
    '''list of any MhdFluid objects.'''
    value_type = MhdFluid
    specie_cls = Specie   # class for making new Specie objects
    ion_mixture_cls = IonMixture   # class for making new IonMixture objects

    def _new_mhd_fluid_list(self, *args, **kw):
        '''return new MhdFluidList (or subclass), with args & kw passed to __init__.
        Subclass might override to also pass info from self into the result.
        '''
        return self._new(*args, **kw)

    def prepend_electron(self, *, electron=UNSET, istart=0, **kw_specie):
        '''return MhdFluidList like self but prepend electron to result.
        electron: UNSET or Specie
            UNSET --> make new electron Specie via Specie.electron().
            Specie --> assert is_electron() then prepend this specie to result.
        istart: None or int. int --> use result[k].with_i(istart + k) for all k.
        '''
        if electron is UNSET:
            electron = self.specie_cls.electron()
        elif not electron.is_electron():
            raise FluidValueError(f'cannot prepend non-electron specie {electron} to list of species')
        return self._new_mhd_fluid_list([electron] + list(self), istart=istart)

    def neutral_list(self, *, unique=False, default=NO_VALUE, istart=0, **kw_specie):
        '''return MhdFluidList of neutral species for elements of fluids in self.
        unique: bool
            whether to give result based on self.unique_elements(), or neutral(f) for f in self.
            False --> result[k] guaranteed to correspond to self[k] for all k.
            True --> ignore fluids without neutral (e.g. electron, or fluid without element),
                     and allow fluids to provide multiple neutrals (e.g. IonMixture with len>1).
        default: NO_VALUE or any value
            if provided, when unique=False and element missing, return default instead of crash.
        istart: None or int.
            int --> use result[k].with_i(istart + k) for all k.
        '''
        if isinstance(default, Element):
            raise NotImplementedError('default=Element not yet supported.')
        if unique:
            elements = self.unique_elements(istart=istart)
        else:
            elements = []
            for f in self:  # use for loop instead of list comprehension for easier debugging if crash.
                el = f.get_element(default=default)  # crashes if element missing and default=NO_VALUE.
                elements.append(el)
        result = [el.neutral(**kw_specie) for el in elements]
        return self._new_mhd_fluid_list(result, istart=istart)

    def ion_list(self, *, q=1, unique=False, default=NO_VALUE, istart=0, **kw_specie):
        '''return MhdFluidList of once-ionized ions for elements of fluids in self.
        q: number, probably int
            charge in elementary charge units. (default: +1)
        unique: bool
            whether to give result based on self.unique_elements(), or ion(f) for f in self.
            False --> result[k] guaranteed to correspond to self[k] for all k.
                      (IonMixtures will remain IonMixtures.)
            True --> ignore fluids without ions (e.g. electron, or fluid without element),
                     and allow fluids to provide multiple ions (e.g. IonMixture with len>1).
        default: bool
            if provided, when unique=False and element missing, return default instead of crash.
        istart: None or int.
            int --> use result[k].with_i(istart + k) for all k.
        '''
        if isinstance(default, Element):
            raise NotImplementedError('default=Element not yet supported.')
        if unique:  # handle IonMixtures separately
            elements = self.unique_elements(istart=istart)
            ions = [el.ion(q=q, **kw_specie) for el in elements]
        else:
            ions = []
            for f in self:
                if isinstance(f, IonMixture):
                    if f.q != q:
                        raise NotImplementedError(f'[TODO] ion_list(q={q}) when q != ion_mixture.q (={f.q})')
                    ions.append(f)
                else:
                    el = f.get_element(default=default)
                    ions.append(el.ion(q=q, **kw_specie))
        return self._new_mhd_fluid_list(ions, istart=istart)

    def saha_list(self, *, istart=0, **kw_specie):
        '''return MhdFluidList of neutral & once-ionized ion for elements of fluids in self.
        Internally, equivalent to: [spec for f in self for spec in f.saha_list()]
        crashes if any fluid is missing an element.
        istart: None or int. int --> use result[k].with_i(istart + k) for all k.
        '''
        result = []
        for f in self:
            result.extend(f.saha_list(**kw_specie))
        return self._new_mhd_fluid_list(result, istart=istart)

    def one_neutral_many_ions(self, *, q=1, unique=False, default=NO_VALUE, istart=0, **kw_specie):
        '''return MhdFluidList of self[0].neutral() then self.ion_list(q=q).
        q, unique, and default all get passed directly to self.ion_list.
        istart: None or int. int --> use result[k].with_i(istart + k) for all k.
        '''
        neutral = self[0].neutral(**kw_specie)
        ions = self.ion_list(q=q, unique=unique, default=default, **kw_specie)
        return self._new_mhd_fluid_list([neutral] + ions, istart=istart)

    # # # MIXING # # #
    def mix(self, fluids, *, append=False, istart=0, **kw_ion_mixture):
        '''return MhdFluidList but with the indicated fluids turned into an IonMixture.
        fluids can be any valid key for self.get which will return an iterable.
            E.g. slice, range, list of indices, list of strings. But not single int or str.
        append: bool
            if True, append mixture instead of putting it at position of first ion that got mixed.
        istart: None or int.
            int --> use result[k].with_i(istart + k) for all k.
        '''
        indices, ions = self.indices(fluids, return_get=True)
        if (getattr(ions, 'ndim', 1) == 0) or (is_iterable(ions) and len(ions) <= 1):
            raise FluidKeyError(f'cannot mix less than 2 fluids; got self.get({fluids!r})=={ions!r}')
        mixed = self.ion_mixture_cls(ions, **kw_ion_mixture)
        i_insert = len(self) if append else min(indices)
        result_without_mix = self.without_i(indices)
        result = result_without_mix.inserted(i_insert, mixed)
        return self._new_mhd_fluid_list(result, istart=istart)

    def mixing(self, *mixes, append=False, istart=0, **kw_ion_mixture):
        '''return MhdFluidList but with the indicated groups of fluids turned into IonMixtures.
        Equivalent: self.mix(mixes[0], **kw).mix(mixes[1], **kw)...

        mixes: iterable of fluids to mix. Each fluid can be any valid key for self.get.
            E.g. slice, range, list of indices, list of strings. But not single int or str.
            CAUTION: mixes are applied in order. Using index-based mixes might produce unexpected results.
        append: bool
            if True, append each mixture instead of putting it at position of first ion that got mixed.
        istart: None or int.
            int --> use result[k].with_i(istart + k) for all k.
        '''
        result = self
        for fluids in mixes:
            result = result.mix(fluids, append=append, istart=istart, **kw_ion_mixture)
        return result

    def mix_heavy_ions(self, m_tol=0.1, *, m_min=5, q=1, append=False, istart=0, **kw_ion_mixture):
        '''return MhdFluidList but grouping heavy ions into as few IonMixtures as possible,
        given the tolerance level for mass deviations within a group, set by m_tol.
        (If any IonMixture would have length==1, leave it as an ion instead.)

        The implementation here takes the naive approach:
            group heavy ions in order, making each group as large as possible.
        This is not necessarily the optimal solution (fewest number of IonMixtures),
            but it produces decent results with minimal compuation / coding effort.
            [TODO] find the actual optimal solution?

        m_tol: number
            maximum allowed relative mass deviation from mean within a group.
            within each group, all ions have |m_ion - mean(m)| < m_tol * mean(m).
            E.g. 0.1 --> all ions in each group are within 10% of that group's mean mass.
        m_min: number
            minimum mass (in amu, i.e. m_H ~= 1) for an ion to be considered "heavy".
            ions with m < m_min are ignored.
        q: number
            charge in elementary charge units. (default: +1)
            ions with different value of q are ignored.
        append: bool
            if True, append each mixture instead of putting it at position of first ion that got mixed.
        istart: None or int.
            int --> use result[k].with_i(istart + k) for all k.
        '''
        heavy_ions = [f for f in self.ions() if (f.m is not None) and (f.m >= m_min)]
        # sort by mass
        heavy_ions = [(f.m, f) for f in heavy_ions]
        heavy_ions.sort()
        heavy_ions = [f for (m, f) in heavy_ions]
        # group into mixtures as large as possible.
        mixtures = []
        while heavy_ions:
            first = heavy_ions.pop(0)
            group = [first]
            msum = first.m
            while heavy_ions:
                ion = heavy_ions[0]
                msum = msum + ion.m
                mmean = msum / (1+len(group))  # 1+ <-- "as if ion was added to group"
                if (ion.m - mmean) > m_tol * mmean:  # new ion would be too massive
                    break
                if (mmean - group[0].m) > m_tol * mmean:  # including new ion would make ions[0] not massive enough
                    break
                # else, this ion belongs in this group.
                group.append(heavy_ions.pop(0))
            mixtures.append(group)
        # remove mixtures with length 1 (IonMixture implementation requires 2+ ions)
        mixtures = [m for m in mixtures if len(m) > 1]
        # convert fluids to strings (prevents crash if multiple mixtures)
        mixtures = [[str(f) for f in m] for m in mixtures]
        # make mixtures
        return self.mixing(*mixtures, append=append, istart=istart, m_tol=m_tol, **kw_ion_mixture)

    # # # GETTING SUBSETS OF VALUES # # #
    def i_species(self):
        '''return indices of Specie fluids in self'''
        return [i for i, f in enumerate(self) if isinstance(f, Specie)]
    def species(self):
        '''return MhdFluidList of Specie fluids in self.'''
        return self[self.i_species()]

    def i_ion_mixtures(self):
        '''return indices of IonMixture fluids in self'''
        return [i for i, f in enumerate(self) if isinstance(f, IonMixture)]
    def ion_mixtures(self):
        '''return MhdFluidList of IonMixture fluids in self.'''
        return self[self.i_ion_mixtures()]

    def ion_mixture_species(self):
        '''return MhdFluidList of Specie fluids from IonMixture fluids in self.'''
        i_mixtures = self.i_ion_mixtures()
        mixtures = [self[i] for i in i_mixtures]
        species = [spec for mix in mixtures for spec in mix.species]
        return self._new_mhd_fluid_list(species)


class ElementHaverList(MhdFluidList):
    '''list of ElementHaver objects.'''
    value_type = ElementHaver


# # # ELEMENTLIST CLASS HOOKUP # # #
# Element is a "virtual subclass" of MhdFluidList, instead of a true subclass,
# to keep the elements.py code focused on single-fluid mhd,
# without worrying about multifluid complications.

ElementHaverList.register(ElementList)  # (see abc.ABC for details)
with binding.to(ElementList):  # assign most of the relevant ElementHandler methods & attributes:
    # cls for making new Specie, IonMixture, ElementList, or MhdFluidList objects:
    ElementList.specie_cls = Specie
    ElementList.ion_mixture_cls = IonMixture
    ElementList.element_list_cls = ElementList
    ElementList.mhd_fluid_list_cls = MhdFluidList
    # stuff that is exactly the same for Element and ElementHaver:
    ElementList.element          = ElementHaverList.element
    ElementList.elements         = ElementHaverList.elements
    ElementList.unique_elements  = ElementHaverList.unique_elements
    ElementList.prepend_electron = ElementHaverList.prepend_electron
    ElementList.neutral_list     = ElementHaverList.neutral_list
    ElementList.ion_list         = ElementHaverList.ion_list
    ElementList.saha_list        = ElementHaverList.saha_list
    ElementList.one_neutral_many_ions = ElementHaverList.one_neutral_many_ions

    # stuff defined here:
    @binding
    def _new_mhd_fluid_list(self, *args, **kw):
        '''return new MhdFluidList, with args & kw passed to __init__.'''
        return self.mhd_fluid_list_cls(*args, **kw)


# # # SPECIELIST # # #

class SpecieList(MhdFluidList):
    '''list of Specie objects.'''
    value_type = Specie


# # # IONMIXTURELIST # # #

class IonMixtureList(MhdFluidList):
    '''list of IonMixture objects'''
    value_type = IonMixture


''' --------------------- Connect stuff that was NotImplemented --------------------- '''

ElementHandler.specie_cls = Specie
ElementHandler.specie_list_cls = SpecieList
Element.specie_cls = Specie
Element.specie_list_cls = SpecieList
IonMixture.specie_list_cls = SpecieList


''' --------------------- SpecialValueSpecifiers --------------------- '''

SPECIES = FluidSpecialValueSpecifier('species', 'SPECIES')
ION_MIXTURES = FluidSpecialValueSpecifier('ion_mixtures', 'ION_MIXTURES')
ION_MIXTURE_SPECIES = FluidSpecialValueSpecifier('ion_mixture_species', 'ION_MIXTURE_SPECIES')



"""
File Purpose: high-level xarray functions. May be especially useful for science.
E.g., gaussian filter, polynomial fit.
"""

import numpy as np
import xarray as xr

from ..imports import ImportFailed
try:
    import scipy.ndimage as scipy_ndimage
except ImportError as err:
    scipy_ndimage = ImportFailed('scipy.ndimage',
            'This module is required for some filtering functions.',
            err=err, locals=locals(), abbrv='scipy_ndimage')
try:
    import scipy.optimize as scipy_optimize
except ImportError as err:
    scipy_optimize = ImportFailed('scipy.optimize',
            'This module is required for some fitting functions.',
            err=err, locals=locals(), abbrv='scipy_optimize')


from .xarray_accessors import pcAccessor
from .xarray_agg_stats import xarray_sum
from .xarray_coords import xarray_is_sorted
from .xarray_dimensions import (
    xarray_promote_dim, xarray_ensure_dims, _paramdocs_ensure_dims,
    xarray_coarsened,
)
from .xarray_grids import xr1d
from .xarray_indexing import xarray_isel, xarray_map
from .xarray_misc import xarray_copy_kw
from ..docs_tools import format_docstring
from ..properties import simple_property
from ..sentinels import UNSET
from ...errors import (
    DimensionValueError,
    InputError, InputConflictError, InputMissingError,
)
from ...defaults import DEFAULTS


''' --------------------- Interpolation --------------------- '''

@pcAccessor.register('interp_inverse', totype='array')
@format_docstring(**_paramdocs_ensure_dims)
def xarray_interp_inverse(array, interpto=None, output=None, *,
                          promote_dims_if_needed=True,
                          assume_sorted=None, assume_sorted_values=None,
                          method=None, kw_interp=None, **interpto_as_kw):
    '''interpolate a DataArray but using the array values as one of the interpolation variables;
    the result is the array of the unused interpolation coordinate.

    Example: if array has dims {{'x', 'y'}} and name 'T', and interpto specifies 'x' and 'T',
        then result will be a DataArray with dims {{'x', 'T'}} and values for 'y'.
        Special case: if interpto specifications are single values,
            the result will be scalar along that key instead of a dimension,
            e.g. if interpto['x'] = 7 (as a opposed to a 1D array like [1,2,3]),
                then result would have coordinate 'x'=7 but not have an 'x' dimension.

    [EFF] note: inefficient if choosing many values along vars other than array.name.
        each result value along those vars corresponds to its own interp call.

    The internal steps are roughly:
        (1) array.interp(all interpto vars except array.name)
        (2) array.assign_coords({{array.name: array}})
        (3) for each index along all interpto vars except array.name:
                tmp = array.isel(index).interp({{array.name: interpto[array.name]}})
                result[index] = tmp[output var]
        [TODO] still need to implement step 3 for 3D+ arrays instead of only 2D or less.

    array: xarray.DataArray
        must have non-None array.name.
    interpto: None or dict
        dictionary of {{var: value or 1D array of values}} to interpolate to.
        Keys must correspond to array.name, and coords for all except 1 dim of array.
        None --> provide interpto dict as kwargs to this function.
    output: None or str
        name for the result variable.
        None --> use the key from array.dims which is missing from interpto (after xarray_ensure_dims).
    promote_dims_if_needed: {promote_dims_if_needed}
    assume_sorted: None or bool
        whether to assume_sorted during step (1),
            i.e. during the initial interp for all interpto vars except array.name
        None --> assume_sorted if xarray_is_sorted.
        True --> assume_sorted without checking. CAUTION: only do this if you're 100% sure!
        False --> don't assume sorted. May be noticeably slower for large arrays.
    assume_sorted_values: None or bool
        whether to assume_sorted during step (3),
            i.e. during the interp using array.name as a coordinate.
        None --> assume_sorted if xarray_is_sorted. check at each index in step 3;
                 if False multiple times in a row, stop checking and just assume False.
        True --> assume_sorted without checking. CAUTION: only do this if you're 100% sure!
        False --> don't assume sorted. May be noticeably slower for large arrays.
    method: None or str
        method to pass to xarray.interp for all interpolations.
        if None, use xarray.interp method default.
    kw_interp: None or dict
        if provided, pass these kwargs to all calls of xarray.interp.
        These will eventually go to the internal interpolator method e.g. from scipy.

    interpto_as_kw: optionally, provide interpto dict as kwargs to this function.
    '''
    # misc. bookkeeping
    if array.name is None:
        raise InputError('interp_inverse expects non-None array.name.')
    if interpto and interpto_as_kw:
        raise InputConflictError('cannot provide both interpto and interpto_as_kw.')
    if interpto_as_kw:
        interpto = interpto_as_kw
    if interpto is None:
        raise InputMissingError('must provide interpto.')
    hard_var = array.name
    easy_vars = set(interpto.keys()) - {hard_var}
    interpto_easy = {k: v for k, v in interpto.items() if k != hard_var}
    interpto_hard = {hard_var: interpto[hard_var]}
    array = xarray_ensure_dims(array, easy_vars, promote_dims_if_needed=promote_dims_if_needed)
    if output is None:
        output = set(array.dims) - easy_vars - {hard_var}
        if len(output) == 0:
            errmsg = ('no acceptable output var found.'
                      f'dims={array.dims}, easy_vars={easy_vars}, hard_var={hard_var!r}')
            raise InputError(errmsg)  # [TODO] clearer debugging suggestions in errmsg.
        if len(output) >= 2:
            errmsg = ('too many acceptable output vars found.'
                      f'dims={array.dims}, easy_vars={easy_vars}, hard_var={hard_var!r}')
            raise InputError(errmsg)  # [TODO] clearer debugging suggestions in errmsg.
        output = output.pop()
    # bookkeeping of kwargs for interp
    if assume_sorted is None:
        assume_sorted = all(xarray_is_sorted(array[v]) for v in easy_vars)
    kw_interp = dict() if kw_interp is None else kw_interp.copy()
    if method is not None:
        if 'method' in kw_interp and method != kw_interp['method']:
            errmsg = f'method={method!r} but kw_interp["method"]={kw_interp["method"]!r}.'
            raise InputConflictError(errmsg)
        # else
        kw_interp['method'] = method
    # (1) intial interpolation
    array = array.interp(interpto_easy, assume_sorted=assume_sorted, **kw_interp)
    if np.any(np.isnan(array)):
        errmsg = ('initial interpolation (step 1 in xarray_interp_inverse) resulted in NaNs.\n'
                  f'This is likely due to some "out of bounds" interpto values ({easy_vars}),\n'
                  'i.e. lower than min or larger than max values of corresponding coords in array.\n'
                  'To fix, adjust requested interpolation coords for interpto vars.\n'
                  '(There is no way to find inverse when step 1 gives nans.)')
        raise DimensionValueError(errmsg)
    # (2) assign
    array = array.assign_coords({hard_var: array})
    # (3) iterating interpolations
    if array.ndim == 1:  # nothing to iterate - hard_var coord corresponds to exactly 1 dim already.
        array = xarray_promote_dim(array, hard_var)
        if assume_sorted_values is None:
            assume_sorted_values = xarray_is_sorted(array)
        array = array.interp(interpto_hard, assume_sorted=assume_sorted_values, **kw_interp)
        result = array[output].drop_vars(output)  # don't need output as a coord, it is result.values.
    elif array.ndim == 2:  # iterate over the 1 easy_var.
        if len(easy_vars) != 1:
            raise NotImplementedError('[TODO] figure out what is happening in this case & handle it.')
        easy_var = easy_vars.pop()
        easy_dim = array[easy_var].dims[0]  # not necessarily easy_var;
            # e.g. if interpto value for easy_var is DataArray with different dim.
        result = []
        if assume_sorted_values is None:
            _prev_assume_sorted = None
        else:
            _assume_sorted = assume_sorted_values
        for i_easy in range(array.sizes[easy_dim]):
            tmp = array.isel({easy_dim: i_easy})
            if assume_sorted_values is None:
                _assume_sorted = xarray_is_sorted(tmp)
                if _prev_assume_sorted == False and _assume_sorted == False:
                    assume_sorted_values = False
                _prev_assume_sorted = _assume_sorted
            tmp = xarray_promote_dim(tmp, hard_var)
            if not _assume_sorted:
                tmp = tmp.sortby(hard_var)
            tmp = tmp.drop_duplicates(hard_var)  # drop duplicates, since interp can't handle them.
            tmp = tmp.interp(interpto_hard, assume_sorted=True, **kw_interp)  # sorted=False handled above.
            tmp = tmp[output].drop_vars(output)  # don't need output as a coord, it is tmp.values.
            result.append(tmp)
        result = xr.concat(result, easy_dim)
    else:
        raise NotImplementedError('[TODO] xarray_interp_inverse with 3D+ array.')
    return result


''' --------------------- Gaussian Filter --------------------- '''

@pcAccessor.register('gaussian_filter', aliases=['blur'])
@format_docstring(**_paramdocs_ensure_dims, default_sigma=DEFAULTS.GAUSSIAN_FILTER_SIGMA)
def xarray_gaussian_filter(array, dim=None, sigma=None, *,
                           promote_dims_if_needed=True, missing_dims='raise',
                           **kw_scipy_gaussian_filter):
    '''returns array after applying scipy.ndimage.gaussian_filter to it.

    array: xarray.DataArray or Dataset
        filters this array, or each data_var in a dataset.
    dim: None or str or iterable of strs
        dimensions to filter along.
        if None, filter along all dims.
    sigma: None, number, or iterable of numbers
        standard deviation for Gaussian kernel.
        if iterable, must have same length as dim.
        if None, will use DEFAULTS.GAUSSIAN_FILTER_SIGMA (default: {default_sigma}).
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}

    additional kwargs go to scipy.ndimage.gaussian_filter.
    '''
    if sigma is None:
        sigma = DEFAULTS.GAUSSIAN_FILTER_SIGMA
    return xarray_map(array, scipy_ndimage.gaussian_filter, sigma, axes=dim,
                      promote_dims_if_needed=promote_dims_if_needed,
                      missing_dims=missing_dims, **kw_scipy_gaussian_filter)


''' --------------------- polyfit --------------------- '''

@pcAccessor.register('polyfit')
@format_docstring(xr_polyfit_docs=xr.DataArray.polyfit.__doc__)
def xarray_polyfit(array, coord, degree, *, stddev=False, full=False, cov=False,
                   eval=False, **kw_polyfit):
    '''returns array.polyfit(coord, degree, **kw_polyfit), after swapping coord to be a dimension, if needed.
    E.g. for an array with dimension 'snap' and associated non-dimension coordinate 't',
        xarray_polyfit(array, 't', 1) is equivalent to array.swap_dims(dict(snap='t')).polyfit('t', 1).

    stddev: bool
        whether to also return the standard deviations of each coefficient in the fit.
        if True, assign the variable 'polyfit_stddev' = diagonal(polyfit_covariance)**0.5,
            mapping the diagonal (across 'cov_i', 'cov_j') to the dimension 'degree'.
            if cov False when stddev True, do not keep_cov in the result.
        Not compatible with full=True.
    full: bool
        passed into polyfit; see below.
    cov: bool
        passed into polyfit; see below.
        Note: if stddev=True when cov=False, still use cov=True during array.polyfit,
            however then remove polyfit_covariance & polyfit_residuals from result.
    eval: bool
        whether to also return the fit, evaluated at array.coords[coord]
        if True, assign the variable 'polyfit_eval' = sum(coeff * coord**degree, 'degree').
            if stddev=True too, assign 'polyfit_eval-stddev' and 'polyfit_eval+stddev' too,
                using similar formula but using coeffs +/- stddev.

    Docs for xr.DataArray.polyfit copied below:
    -------------------------------------------
    {xr_polyfit_docs}
    '''
    array = xarray_promote_dim(array, coord)
    if stddev and full:
        raise InputConflictError('stddev=True incompatible with full=True.')
    cov_input = cov
    if stddev:
        cov = True
    result = array.polyfit(coord, degree, full=full, cov=cov, **kw_polyfit)
    if stddev:
        result = xarray_assign_polyfit_stddev(result, keep_cov=cov_input)
    if eval:
        evaled = xarray_polyfit_eval(result, at=array.coords[coord], to_dataset=True)
        result = result.assign(evaled)
    return result

@pcAccessor.register('assign_polyfit_stddev', totype='dataset')
def xarray_assign_polyfit_stddev(dataset, *, keep_cov=True):
    '''assign polyfit stddev to dataset['polyfit_stddev'], treating dataset like a result of polyfit.
    These provide some measure of "goodness of fit"; smaller stddev means better fit.

    Specifically, stddev[k] = (covariance matrix)[k,k]**0.5 for k in range(len(dataset['degree']));
        one might quote +-stddev[k] as the error bar for the coefficient at degree=dataset['degree'][k].

    dataset: xarray.Dataset
        dataset to use for calculating polyfit_stderr and in which to assign the result.
        must contain variable 'polyfit_covariance' and dimension 'degree'.
    keep_cov: bool
        whether to keep the 'polyfit_covariance' and 'polyfit_residuals' vars in the result.

    The original dataset will not be altered; a new dataset will be returned.
    '''
    cov = dataset['polyfit_covariance']
    degree = dataset['degree']
    ndeg = len(degree)
    stddev = [cov.isel(cov_i=k, cov_j=k)**0.5 for k in range(ndeg)]
    stddev = xr.concat(stddev, 'degree').assign_coords({'degree': degree})
    result = dataset.assign(polyfit_stddev=stddev)
    if not keep_cov:
        result = result.drop_vars(['polyfit_covariance', 'polyfit_residuals'])
    return result

@pcAccessor.register('polyfit_eval', totype='dataset')
def xarray_polyfit_eval(dataset, at=None, *, to_dataset=False):
    '''evaluate the polyfit in dataset at these coordinates.
    E.g. dataset has degree=2,1,0, polyfit_coefficients=7.5, -3.2, 0.4
        --> result == 7.5 * at**2 - 3.2 * at + 0.4

    if input dataset contains 'polyfit_stddev', instead returns a dataset with:
        'polyfit_eval': result (see above)
        'polyfit_eval-stddev': result when using polyfit_coefficients - polyfit_stddev.
        'polyfit_eval+stddev': result when using polyfit_coefficients + polyfit_stddev.

    at: None or array
        if provided, evaluate polyfit dataset here.
        None --> input MUST be a dataset which already has 'polyfit_eval';
            instead of computing terms, just return relevant info which exists in dataset,
            i.e. 'polyfit_eval', 'polyfit_eval-stddev', and/or 'polyfit_eval+stddev'.
    to_dataset: bool
        whether to return dataset even if input does not contain 'polyfit_stddev'
        True --> result will always be a dataset.
    '''
    if at is None:
        if 'polyfit_eval' not in dataset:
            return InputMissingError("when at=None, dataset must contain 'polyfit_stddev'.")
        relevant = ['polyfit_eval']
        for v in ['polyfit_eval-stddev', 'polyfit_eval+stddev']:
            if v in dataset:
                relevant.append(v)
        return dataset[relevant]
    coeffs = dataset['polyfit_coefficients']
    if 'polyfit_stddev' in dataset:
        stddev = dataset['polyfit_stddev']
        coeffs = xr.Dataset({'polyfit_eval': coeffs,
                             'polyfit_eval-stddev': coeffs - stddev,
                             'polyfit_eval+stddev': coeffs + stddev})
    elif to_dataset:
        coeffs = xr.Dataset({'polyfit_eval': coeffs})
    terms = coeffs * at**dataset['degree']
    return xarray_sum(terms, 'degree')

@pcAccessor.register('coarsened_polyfit')
@format_docstring(xr_polyfit_docs=xr.DataArray.polyfit.__doc__)
def xarray_coarsened_polyfit(array, coord, degree, window_len, *,
                             dim_coarse='window', keep_coord='middle',
                             assign_coarse_coords=True,
                             boundary=UNSET, side=UNSET,
                             stride=UNSET, fill_value=UNSET, keep_attrs=UNSET,
                             **kw_polyfit
                             ):
    '''returns result of coarsening array, then polyfitting along the fine dimension, in each window.
    E.g., make windows of length 10 along 't', then polyfit each window along 't',
    then concat the results from each window, along dim_coarse (default: 'window').

    coord: str
        coordinate to polyfit along.
    degree: int
        degree of polynomial to fit.
    window_len: int or None
        length of window to coarsen over.
        None --> polyfit without coarsening; equivalent to window_len = len(array.coords[coord])
    dim_coarse: str, default 'window'
        name of coarse dimension; the i'th value here corresponds to the i'th window.
    keep_coord: False or str in ('left', 'right', 'middle')
        along the dim_coarse dimension, also provide some of the original coord values.
        'left' --> provide the left-most value in each window.
        'middle' --> provide the middle value in each window.
        'right' --> provide the right-most value in each window.
        False --> don't provide any of the original coord values.
        if not False, result will swap dims such that coord is a dimension instead of dim_coarse.
    assign_coarse_coords: bool or coords
        coords to assign along the dim_coarse dimension.
        True --> use np.arange.
        False --> don't assign coords.
    boundary, side: UNSET or value
        if provided (not UNSET), pass this value to coarsen().
    stride, fill_value, keep_attrs: UNSET or value
        if provided (not UNSET), pass this value to construct().

    additional **kw are passed to polyfit.

    Docs for xr.DataArray.polyfit copied below:
    -------------------------------------------
    {xr_polyfit_docs}
    '''
    # bookkeeping
    if keep_coord not in ('left', 'middle', 'right', False):
        raise InputError(f'invalid keep_coord={keep_coord!r}; expected "left", "middle", "right", or False.')
    # if window_len is None or <1, don't coarsen at all.
    if window_len is None:
        return xarray_polyfit(array, coord, degree, **kw_polyfit)
    # coarsen
    coarsened = xarray_coarsened(array, coord, window_len,
                                dim_coarse=dim_coarse,
                                assign_coarse_coords=assign_coarse_coords,
                                boundary=boundary, side=side,
                                stride=stride, fill_value=fill_value, keep_attrs=keep_attrs)
    # bookkeeping
    n_windows = len(coarsened[dim_coarse])
    if n_windows < 1:
        errmsg = f'coarsened array has n_windows={n_windows} < 1; cannot polyfit.'
        raise DimensionValueError(errmsg)
    # polyfitting
    promoted = []
    for i_window in range(n_windows):
        prom = xarray_promote_dim(coarsened.isel({dim_coarse: i_window}), coord)
        promoted.append(prom)
    polyfits = []
    for arr in promoted:
        pfit = xarray_polyfit(arr, coord, degree, **kw_polyfit)
        polyfits.append(pfit)
    if keep_coord:
        results = []
        for i_window, (arr, prom) in enumerate(zip(polyfits, promoted)):  # i_window just for debugging
            i_keep = {'left': 0, 'middle': 0.5, 'right': -1}[keep_coord]
            # isel from coords[coord] instead of prom, to ensure associated coords are included too,
            #   e.g. t & snap are associated --> this will keep t & snap in the result.
            # if i_keep = 0.5, it is handled by xarray_isel fractional indexing.
            keep = xarray_isel(prom.coords[coord], {coord: i_keep})
            here = arr.assign_coords({coord: keep})
            results.append(here)
    else:
        results = polyfits
    result = xr.concat(results, dim_coarse)
    if keep_coord:
        result = xarray_promote_dim(result, coord)
    return result


''' --------------------- curve fit --------------------- '''

_paramdocs_curve_fit = {
    'array': '''xarray.DataArray or Dataset
        data to fit.
        Currently, Dataset allowed only if it has 'mean' and 'std' data_vars, when `stddev=True`,
            in which case will sample the implied gaussians (via np.random.normal),
            N=`werr_samples` times, performing N fits to f,
            reporting the mean and stddev of each fit param across all N fits, and
            ignoring scipy standard deviation info about params from each individual fit.''',
    'werr_samples': '''int
        number of fits to do if `array` is a Dataset with 'mean' and 'std' vars, when `stddev=True`,
            in which case result will tell mean and stddev of each fit param across all N fits,
            and ignore scipy standard deviation info about params from each individual fit.
        (Implemented this because default scipy linear least squares fitting with errorbars
            just weights each point's important by inverse of error bar,
            which highly favors points with small errors.
        That default does NOT correspond to the results of "repeating the experiment" N times,
            where "the experiment" is gathering data then fitting,
            and then asking "what is the mean and stddev of fit params across all N experiments?".
        However, using werr_samples DOES correspond to "repeating the experiment" N times.)''',
    'werr_seed': '''None or any object, default 0
        `np.random.seed(werr_seed)` beforehand, if doing werr_samples (with Dataset `array`).
        Default 0 ensures reproducible results.
        None --> don't call np.random.seed beforehand. Will give different results each time.''',
    'pnames': '''None or list of str
        names of params. If provided, 'param' coord will be assigned these names.''',
    'pbounds': '''None or list of [None, callable, or 2-tuple of value, None, or callable]
        bounds for each parameter. Provide `pbounds` or `bounds`, but not both.
        None --> no bounds provided.
        Each bound can be:
            callable --> call as bound(array, dim) (after doing array.pc.ensure_dims(dim)).
            None --> use (-np.inf, np.inf).
            2-tuple --> (lower, upper).
                callable --> use lower(array, dim) / upper(array, dim)
                None --> use -np.inf / np.inf.''',
    'bounds': '''UNSET or (list of lower bounds, list of upper bounds)
        bounds, formatted as expected by scipy curve_fit.
        Provide `pbounds` or `bounds`, but not both.''',
}



@pcAccessor.register('curve_fit')
@format_docstring(**_paramdocs_ensure_dims, **_paramdocs_curve_fit)
def xarray_curve_fit(array, f, dim, *, stddev=True, werr_samples=1000, werr_seed=0,
                     promote_dims_if_needed=True,
                     pnames=None, pbounds=None, bounds=UNSET, **kw_curve_fit):
    '''scipy.optimize.curve_fit(f, xdata=array[dim], ydata=array).
    Except, iterate over all other dims in array.
        E.g. arr.curve_fit('t', f) for arr with 't' and 'fluid' dims
        --> result reduces 't' dim but retains 'fluid' dim.

    array: {array}
    dim: str
        dim to fit along.
    f: callable like f(x, param1, param2, ...)
        function to fit.
    stddev: bool
        whether to include data_var 'stddev' telling standard deviation of the fit.
    werr_samples: {werr_samples}
    werr_seed: {werr_seed}
    promote_dims_if_needed: {promote_dims_if_needed}
    pnames: {pnames}
    pbounds: {pbounds}
    bounds: {bounds}

    additional kwargs go to scipy.optimize.curve_fit.

    returns xarray.Dataset with data_vars:
        params: iterable along 'param' dimension of the parameters of the fit.
            e.g. call f with f(x, *result['params'].values) if 1D array.
            e.g. f(x, *result['params'].isel(fluid=0).values) if 2D array.
        stddev: standard deviation of each parameter's fit.
    '''
    array_orig = array
    if isinstance(array, xr.Dataset):
        if stddev != True:
            raise InputError('xarray_curve_fit(Dataset), only implemented when stddev=True.')
        if 'mean' not in array or 'std' not in array:
            errmsg = ('xarray_curve_fit(Dataset) only implemented when input has "mean" and "std" data_vars;'
                     f' got data_vars: {list(array.data_vars)}')
            raise InputError(errmsg)
        if werr_seed is not None:
            np.random.seed(werr_seed)
        # sample the implied gaussians N times, then fit each sample.
        mean = array['mean']
        std = array['std']
        samples = np.random.normal(mean, std, size=(werr_samples,) + mean.shape)
        array = xr.DataArray(samples, dims=('__werr_sample__', *mean.dims),
                             coords=mean.coords)
        werrmode = True
        stddev = False  # don't compute the stddevs from scipy for individual fits, below!
    else:
        werrmode = False
    array = xarray_ensure_dims(array, dim, promote_dims_if_needed=promote_dims_if_needed)
    # bounds
    if pbounds is not None and bounds is not UNSET:
        raise InputConflictError('provide pbounds or bounds, but not both.')
    if pbounds is not None:
        bounds = [b(array, dim) if callable(b) else b for b in pbounds]
        bounds = [(-np.inf, np.inf) if b is None else b for b in bounds]
        bounds = [((l(array, dim) if callable(l) else l),
                   (u(array, dim) if callable(u) else u)) for l, u in bounds]
        bounds = [((-np.inf if l is None else l),
                   (np.inf if u is None else u)) for l, u in bounds]
        bounds = tuple(zip(*bounds))
    if bounds is not UNSET:
        kw_curve_fit['bounds'] = bounds
    # back to non-bound-stuff
    other = list(set(array.dims) - {dim})
    sizes = [array.sizes[d] for d in other]
    result = None
    stds = None
    for index in np.ndindex(*sizes):
        sub = array.isel(dict(zip(other, index)))
        x = sub[dim].values
        y = sub.values
        params, cov = scipy_optimize.curve_fit(f, x, y, **kw_curve_fit)
        if result is None:  # first time --> make full result with proper shape
            result = np.empty([*sizes, len(params)], dtype=params.dtype)
        result[index] = params
        if stddev:
            if stds is None:
                stds = np.empty([*sizes, len(params)], dtype=params.dtype)
            stds[index] = np.sqrt(np.diag(cov))
    # bookkeeping & putting it all together
    if werrmode:
        assert other[0] == '__werr_sample__'  # 0'th dim in "other" is the sample dim.
        stds = result.std(axis=0)
        result = result.mean(axis=0)
        stddev = True  # do include these stddevs in the final result, below!
        other = other[1:]
    kw = xarray_copy_kw(array_orig, dims=other)
    kw['dims'] = list(kw['dims']) + ['param']
    if pnames is not None:
        kw['coords'] = {**kw.get('coords', {}), 'param': pnames}
    result = xr.DataArray(result, **kw)
    result = xr.Dataset({'params': result})
    if stddev:
        stds = xr.DataArray(stds, **kw)
        result = result.assign(stddev=stds)
    return result

@pcAccessor.register('curve_eval')
def xarray_curve_eval(params, f, xdata, *, stddev=False):
    '''evaluate a curve fit result (params) for this function at these xdata.
    [EFF] note: well-vectorized f could equivalently do:
        f(xdata, *params.transpose('param', ...))
        and that would probably be much faster, if it is an available option.

    params: xarray.DataArray or xarray.Dataset
        curve_fit result, or result['params']. Must have 'param' dimension.
        if Dataset, will internally use params['params'].
    f: callable like f(x, param1, param2, ...)
        function which was fit / to be evaluated.
    xdata: 1D xarray.DataArray
        x values at which to evaluate the fit.
    stddev: bool
        whether to instead return Dataset with data_vars 'eval+std', 'eval', and 'eval-std',
        telling f(xdata, *(params+std)), f(xdata, *params), and f(xdata, *(params-std)).
        Fails if params is not a Dataset with 'stddev' data_var.
    '''
    if stddev:
        if not isinstance(params, xr.Dataset) or 'stddev' not in params:
            raise InputError('stddev=True but params is not a Dataset with "stddev" data_var.')
        stds = params['stddev']
    if isinstance(params, xr.Dataset):
        params = params['params']
    if not isinstance(xdata, xr.DataArray) or xdata.ndim != 1:
        raise InputError('xdata must be a 1D xarray.DataArray.')
    # misc. bookkeeping:
    other = list(set(params.dims) - {'param'})
    sizes = [params.sizes[d] for d in other]
    kw = xarray_copy_kw(params, dims=other)
    kw['dims'] = list(kw['dims']) + [xdata.dims[0]]
    kw_xdata = xarray_copy_kw(xdata)
    kw['coords'] = {**kw.get('coords', {}), **kw_xdata.get('coords', {})}
    # computing result:
    # [TODO] encapsulate "apply function looping across irrelevant dims" elsewhere...
    if stddev:
        result = np.empty([3, *sizes, len(xdata)], dtype=params.dtype)
        for index in np.ndindex(*sizes):
            ii = dict(zip(other, index))
            params_ii = params.isel(ii)
            stds_ii = stds.isel(ii)
            result[0][index] = f(xdata, *(params_ii+stds_ii).values)
            result[1][index] = f(xdata, *(params_ii).values)
            result[2][index] = f(xdata, *(params_ii-stds_ii).values)
        evaled = {
            'eval+std': xr.DataArray(result[0], **kw),
            'eval':     xr.DataArray(result[1], **kw),
            'eval-std': xr.DataArray(result[2], **kw),
        }
        kw.pop('dims', None)  # Dataset doesn't want you to enter 'dims' explicitly.
        kw.pop('name', None)  # Dataset doesn't want you to enter 'name'.
        result = xr.Dataset(evaled)
    else:
        result = np.empty([*sizes, len(xdata)])
        for index in np.ndindex(*sizes):
            sub = params.isel(dict(zip(other, index)))
            result[index] = f(xdata, *sub.values)
        result = xr.DataArray(result, **kw)
    # bookkeeping & putting it all together
    return result

@format_docstring(**_paramdocs_ensure_dims, **_paramdocs_curve_fit)
class XarrayCurveFitter():
    '''class for helping with curve fitting.
    Not intended for direct use; subclass must define f:
        f: function to fit. callable like f(x, param1, param2, ...)

    (For direct curve fitting from array, with arbitrary function f,
        without implementing subclass, consider array.pc.curve_fit(f, dim) instead.)

    array: {array}
    dim: str
        dim to fit along.
    stddev: bool
        whether to include data_var 'stddev' telling standard deviation of the fit.
        (Internally, stored inside self.kw_curve_fit)
    werr_samples: {werr_samples}
        (Internally, stored inside self.kw_curve_fit)
    werr_seed: {werr_seed}
    promote_dims_if_needed: {promote_dims_if_needed}
    pnames: UNSET or {pnames}
        UNSET --> use cls.pnames (default: None)
    pbounds: UNSET or {pbounds}
        UNSET --> use cls.pbounds (default: None)
    bounds: {bounds}

    additional kwargs go to xarray_curve_fit, then scipy.optimize.curve_fit.
    '''
    pnames = None
    pbounds = None

    def __init__(self, array, dim, *, promote_dims_if_needed=True,
                 pnames=UNSET, pbounds=UNSET, bounds=UNSET, **kw_curve_fit):
        array = xarray_ensure_dims(array, dim, promote_dims_if_needed=promote_dims_if_needed)
        self.array = array
        self.dim = dim
        if pnames is not UNSET: self.pnames = pnames
        if pbounds is not UNSET: self.pbounds = pbounds
        self.bounds = bounds
        self.kw_curve_fit = kw_curve_fit

    def f(self, x, *params):
        '''function to fit. callable like f(x, param1, param2, ...).
        [Not implemented here; subclass should implement]
        '''
        raise NotImplementedError(f'{type(self).__name__}.f')

    fitted = simple_property('_fitted', default=None,
        doc='''result of latest call to self.fit().
        None if never called self.fit(), or if crashed before finishing self.fit().''')

    @property
    def params(self):
        '''alias to self.fitted['params'], the params from latest call to self.fit.
        crash with helpful message if self.fitted doesn't exist.
        '''
        if self.fitted is None:
            raise AttributeError('params not available; call self.fit() first.')
        return self.fitted['params']

    @property
    def xdata(self):
        '''alias to self.array[self.dim]'''
        return self.array[self.dim]

    def fit(self, *, stddev=UNSET):
        '''curve_fit to ydata = self.array, xdata = self.array[self.dim].
        Remembers result in self.fitted. Returns self.fitted.

        stddev: UNSET or bool
            whether to include data_var 'stddev' telling standard deviation of the fit.
            UNSET --> use value from self.kw_curve_fit, else default of xarray_curve_fit.
        '''
        self.fitted = None  # clear old result, if one exists.
        kw = dict(pnames=self.pnames, pbounds=self.pbounds, bounds=self.bounds)
        kw.update(self.kw_curve_fit)
        if stddev is not UNSET: kw['stddev'] = stddev
        result = xarray_curve_fit(self.array, self.f, self.dim, **kw)
        self.fitted = result
        return self.fitted

    def eval(self, xdata=UNSET, params=UNSET, stddev=False):
        '''evaluate curve fit result (params) at these xdata.
        Equivalent: xarray_curve_eval(params, self.f, xdata)

        xdata: UNSET, 1D xarray.DataArray, or other 1D array-like
            x values at which to evaluate the fit.
            UNSET --> use self.xdata.
            non-xarray 1D array-like --> convert to 1D DataArray via xr1d(xdata, self.dim)
        params: UNSET or values of params from a fit.
            UNSET --> use self.fitted
        stddev: bool
            whether to instead return Dataset with data_vars 'eval+std', 'eval', and 'eval-std',
            telling f(xdata, *(params+std)), f(xdata, *params), and f(xdata, *(params-std)).
            Fails if params is not a Dataset with 'stddev' data_var.

        [EFF] note: if self.f is well-vectorized, it is equivalent (when stddev=False) and faster to do:
            self.f(xdata, *params.transpose('param', ...))
        '''
        if xdata is UNSET: xdata = self.xdata
        elif not isinstance(xdata, xr.DataArray):
            xdata = xr1d(xdata, self.dim)
        if params is UNSET: params = self.fitted
        return xarray_curve_eval(params, self.f, xdata, stddev=stddev)

    def __repr__(self):
        contents = []
        if self.array is not None:
            contents.append(f'array=<{type(self.array).__name__} at {hex(id(self.array))}>')
        if self.fitted is not None:
            contents.append(f'fitted=<{type(self.fitted).__name__} at {hex(id(self.fitted))}>')
        return f'{type(self).__name__}({", ".join(contents)})'


@pcAccessor.register('line_fitter')
@format_docstring(**_paramdocs_ensure_dims, **_paramdocs_curve_fit)
class XarrayLineFitter(XarrayCurveFitter):
    '''XarrayCurveFitter with f a line: f(x, slope, intercept) = slope * x + intercept.

    array: {array}
    dim: str
        dim to fit along.
    stddev: bool
        whether to include data_var 'stddev' telling standard deviation of the fit.
        (Internally, stored inside self.kw_curve_fit)
    werr_samples: {werr_samples}
        (Internally, stored inside self.kw_curve_fit)
    werr_seed: {werr_seed}
         (Internally, stored inside self.kw_curve_fit)
    promote_dims_if_needed: {promote_dims_if_needed}
    pnames: UNSET or {pnames}
        UNSET --> use cls.pnames (default: None)
    pbounds: UNSET or {pbounds}
        UNSET --> use cls.pbounds (default: None)
    bounds: {bounds}

    additional kwargs go to xarray_curve_fit, then scipy.optimize.curve_fit.
    '''
    @staticmethod
    def f(x, slope, intercept):
        '''function to fit: just a simple line. f(x, slope, intercept) = slope * x + intercept.'''
        return slope * x + intercept

    pnames = ['slope', 'intercept']


@pcAccessor.register('line_fit', aliases=['linear_fit'])
@format_docstring(**_paramdocs_ensure_dims, **_paramdocs_curve_fit)
def xarray_line_fit(array, dim, *, pnames=UNSET, pbounds=UNSET, **kw_curve_fitter):
    '''returns result of xarray_curve_fit with f a line:
        f(x, slope, intercept) = slope * x + intercept.

    array: {array}
    dim: str
        dim to fit along.
    stddev: bool
        whether to include data_var 'stddev' telling standard deviation of the fit.
    werr_samples: {werr_samples}
    werr_seed: {werr_seed}
    promote_dims_if_needed: {promote_dims_if_needed}
    pnames: UNSET or {pnames}
        UNSET --> use pnames = ['slope', 'intercept']
    pbounds: UNSET or {pbounds}
        UNSET --> use pbounds = None
    bounds: {bounds}

    additional kwargs go to XarrayCurveFitter, then xarray_curve_fit, then scipy.optimize.curve_fit.
    '''
    fitter = XarrayLineFitter(array, dim, pnames=pnames, pbounds=pbounds, **kw_curve_fitter)
    return fitter.fit()



"""
File Purpose: handling imports related to tfbi.
"""

from ...defaults import DEFAULTS
from ...tools import ImportFailed

# if SymSolver or tfbi_theory imports fail, some tfbi methods will fail,
#  but not all methods --> still useful to load this addon.
try:
    import SymSolver
except ImportError as err:
    SymSolver = ImportFailed("SymSolver", err=err, locals=locals(), abbrv='SymSolver')
try:
    import tfbi_theory
except ImportError as err:
    tfbi_theory = ImportFailed("tfbi_theory", err=err, locals=locals(), abbrv='tfbi_theory')

if (DEFAULTS.ADDONS.LOAD_TFBI == True) and any(isinstance(x, ImportFailed) for x in (SymSolver, tfbi_theory)):
    errmsg = ("Failed to import SymSolver or tfbi_theory. To disable this error, "
              "Set PlasmaCalcs.defaults.DEFAULTS.LOAD_TFBI to 'attempt' or False.")
    raise ImportError(errmsg)



"""
File Purpose: methods related to dims for plotting.

See also: DEFAULTS.DIMS_INFER
"""

from ...errors import PlottingAmbiguityError, DimensionKeyError

from ...tools import (
    simple_tuple_property,
    NO_VALUE,
)
from ...defaults import DEFAULTS

''' --------------------- inferring dimensions --------------------- '''

def infer_movie_dim(dims, dim=None, *, exclude=[], fail_ok=True):
    '''infer the dimension which movie frames should index.
    (if dim is provided, just return dim, after ensuring dim in dims)

    dims: iterable of strings
        all possible dims. Probably array.dims, if trying to infer from an array.
    dim: None or str
        the dimension which frames will index.
        if None, infer it, if possible:
            if len(dims)==1, use the only dim remaining from dims.
            otherwise, use the first dim from DEFAULTS.PLOT.DIMS_INFER['t'] which appears in dims, if possible.
            otherwise, failed to infer a dim; return None.
    exclude: iterable
        dims to exclude from dims, before inferring dim.
        E.g. if dims=('x', 'y', 'z', 't') and exclude=('x', 'y'), then infer from dims=('z', 't').
        Useful e.g. if it is already known which dims are used for x & y axes of the plot.
        Exclude can contain any values, e.g. exclude=('x', None) is fine.
    fail_ok: bool
        controls behavior when failing to infer dim:
        True --> return None
        False --> raise PlottingAmbiguityError
    '''
    dims = tuple(dims)  # compatible with generator iterables
    if dim is not None:
        if dim not in dims:
            raise DimensionKeyError(f'dim={dim!r} not found in dims={dims!r}')
        return dim
    # infer dim from dims
    dims = list(set(dims) - set(exclude))
    if len(dims)==1:
        return dims[0]
    for d in DEFAULTS.PLOT.DIMS_INFER['t']:
        if d in dims:
            return d
    # failed to infer dim.
    if fail_ok:
        return None
    raise PlottingAmbiguityError(f'failed to infer movie dim from dims={dims!r}')

def infer_xy_dims(dims, x=None, y=None, *, exclude=[], fail_ok=False):
    '''infer x and y dims from dims, if not provided.
    (if x and y are provided, just return x,y, after ensuring x,y in dims)

    dims: iterable of strings
        all possible dims. Probably array.dims, if trying to infer from an array.
    x: None or str
        the dimension which will index the plot's x axis.
        if None, infer it, if possible:
            if y is known, first set dims = dims without y.
            if len(dims)==1, use x=dims[0]
            otherwise, use the first dim from DEFAULTS.PLOT.DIMS_INFER['x'] which appears in dims, if possible.
            otherwise, try to infer y; might be able to infer x while inferring y.
    y: None or str
        the dimension which will index the plot's y axis.
        if None, infer it, if possible:
            if x is known, first set dims = dims without x.
            if len(dims)==1, use the only dim remaining from dims.
            otherwise, use the first dim from DEFAULTS.PLOT.DIMS_INFER['y'] which appears in dims, if possible.
            otherwise, if len(dims)==2 and x is None, use x, y = dims
    exclude: iterable
        dims to exclude from dims, before inferring x,y.
        E.g. if dims=('x', 'y', 'z', 't') and exclude=('x', 'y'), then infer from dims=('z', 't').
        Useful e.g. if it is already known which dims are used for x & y axes of the plot.
        Exclude can contain any values, e.g. exclude=('x', None) is fine.
    fail_ok: bool
        controls behavior when failing to infer dim:
        True --> return None for that dim
        False --> raise PlottingAmbiguityError
    '''
    dims = tuple(dims)  # compatible with generator iterables
    if (x is not None) and (x not in dims):
        raise DimensionKeyError(f'x={x!r} not found in dims={dims!r}')
    if (y is not None) and (y not in dims):
        raise DimensionKeyError(f'y={y!r} not found in dims={dims!r}')
    if (x is not None) and (y is not None):
        return (x, y)
    # infer x and/or y from dims
    dims = list(set(dims) - set(exclude))
    if x is None:
        if y is not None:
            dims = list(set(dims) - {y})
        if len(dims)==1:
            x = dims[0]
        if x is None:  # x is still None
            for dim in DEFAULTS.PLOT.DIMS_INFER['x']:
                if dim in dims:
                    x = dim
                    break
    if y is None:
        if x is not None:
            dims = list(set(dims) - {x})
        if len(dims)==1:
            y = dims[0]
        if y is None:  # y is still None
            for dim in DEFAULTS.PLOT.DIMS_INFER['y']:
                if dim in dims:
                    y = dim
                    break
        if y is None:  # y is still None
            if len(dims)==2 and x is None:
                x, y = dims
    # return result; first, crash if failed but not fail_ok.
    if not fail_ok:
        if (x is None) and (y is None):
            raise PlottingAmbiguityError(f'failed to infer x,y from dims={dims!r}')
        if x is None:
            raise PlottingAmbiguityError(f'failed to infer x from dims={dims!r}, y={y!r}')
        if y is None:
            raise PlottingAmbiguityError(f'failed to infer y from dims={dims!r}, x={x!r}')
    return (x, y)

def infer_xyt_dims(dims, *, x=None, y=None, t=None, exclude=[], xy_fail_ok=False, t_fail_ok=True):
    '''infer x, y, and t dims from dims, if not provided.
    (if x, y, and t are provided, just return x,y,t, after ensuring x,y,t in dims)

    dims: iterable of strings
        all possible dims. Probably array.dims, if trying to infer from an array.
        This function expects len(dims)==2 or 3, otherwise might behave strangely.
    x: None or str
        the dimension which will index the plot's x axis.
    y: None or str
        the dimension which will index the plot's y axis.
    t: None or str
        the dimension which will index the movie's frames.
    exclude: iterable
        dims to exclude from dims, before inferring x,y,t.
    xy_fail_ok, t_fail_ok: bool
        controls behavior when failing to infer a dim:
        True --> return None for that dim
        False --> raise PlottingAmbiguityError
        t_fail_ok corresponds to t; xy_fail_ok corresponds to x,y.

    see also: infer_xy_dims, infer_movie_dim, DEFAULTS.PLOT.DIMS_INFER
    '''
    dims = tuple(set(dims) - set(exclude))
    if len(dims) in (2,3):
        t = infer_movie_dim(dims, t, fail_ok=True)
    else:
        t = infer_movie_dim(dims, t, exclude=[x, y], fail_ok=True)
    x, y = infer_xy_dims(dims, x, y, exclude=[t], fail_ok=xy_fail_ok)
    if t is None:
        t = infer_movie_dim(dims, t, exclude=[x, y], fail_ok=t_fail_ok)
    return (x, y, t)


class PlotDimsMixin():
    '''adds methods for inferring plot dims, and properties for plot dims.
    subclasses should define plot_dims_attrs = dict(
        x = name of attr which stores the un-inferred dim for plot x axis
        y = name of attr which stores the un-inferred dim for plot y axis
        t = name of attr which stores the un-inferred dim for plot t axis (i.e. frames in a movie)
        dims = name of attr which stores the list of all possible dims, from which other dims might be inferred.
        t_necessary_if = name of attr which stores the condition under which t does not need to be inferred.
    ). If not all of these are provided, some (or all) InferrablePlotDimsMixin methods will crash.

    defines properties:
        x_plot_dim, y_plot_dim, t_plot_dim:
            infer the related dim (x,y,t, respectively),
            using values at plot_dims_attrs, and DEFAULTS.PLOT.DIMS_INFER.
            setting x_plot_dim=value also sets self.{plot_dims_attrs['x']}=value, etc.
        xy_plot_dims:
            (x_plot_dim, y_plot_dim); setting value also sets self.x_plot_dim and y_plot_dim.
        xyt_plot_dims:
            similar to xy_plot_dims, but for (x_plot_dim, y_plot_dim, t_plot_dim).

    For an example of using this Mixin, see XarrayImage.
    '''
    def _get_plot_dims_attr_value(self, x, default=NO_VALUE):
        '''get value of self.{plot_dims_attrs[x]}, where x = 'x', 'y', 't', 'dims', or 't_necessary_if'.'''
        attr = self.plot_dims_attrs[x]
        return getattr(self, attr) if default is NO_VALUE else getattr(self, attr, default)
    def _set_plot_dims_attr_value(self, x, value):
        '''set value of self.{plot_dims_attrs[x]}, where x = 'x', 'y', 't', 'dims', or 't_necessary_if'.'''
        attr = self.plot_dims_attrs[x]
        setattr(self, attr, value)

    @property
    def x_plot_dim(self):
        '''the dimension which is actually used to index the plot's x axis.
        setting self.x_plot_dim=value also sets self.{plot_dims_attrs['x']}=value.
        getting self.x_plot_dim will get self.{plot_dims_attrs['x']} if it is not None;
            otherwise, infer from self.{plot_dims_attrs['dims']}, self.{plot_dims_attrs['t']}, and self.{plot_dims_attrs['y']}.
        '''
        # [TODO][EFF] cache inferred value? (Only worthwhile if profiling reveals this is slow.)
        result = self._get_plot_dims_attr_value('x')
        if result is not None:
            return result
        x, y, t = self.infer_xyt_dims(xy_fail_ok=False, t_fail_ok=True)
        return x
    @x_plot_dim.setter
    def x_plot_dim(self, value):
        self._set_plot_dims_attr_value('x', value)

    @property
    def y_plot_dim(self):
        '''the dimension which is actually used to index the plot's y axis.
        setting self.y_plot_dim=value also sets self.{plot_dims_attrs['y']}=value.
        getting self.y_plot_dim will get self.{plot_dims_attrs['y']} if it is not None;
            otherwise, infer from self.{plot_dims_attrs['dims']}, self.{plot_dims_attrs['t']}, and self.{plot_dims_attrs['x']}.
        '''
        result = self._get_plot_dims_attr_value('y')
        if result is not None:
            return result
        x, y, t = self.infer_xyt_dims(xy_fail_ok=False, t_fail_ok=True)
        return y
    @y_plot_dim.setter
    def y_plot_dim(self, value):
        elf._set_plot_dims_attr_value('y', value)

    @property
    def t_plot_dim(self):
        '''the dimension which is actually used to index the movie's frames (i.e. the "time axis").
        setting self.t_plot_dim=value also sets self.{plot_dims_attrs['t']}=value.
        getting self.t_plot_dim will get self.{plot_dims_attrs['t']} if it is not None;
            otherwise, infer from self.{plot_dims_attrs['dims']}, self.{plot_dims_attrs['x']}, and self.{plot_dims_attrs['y']}
            if self.{plot_dims_attrs['t_necessary_if']}(), result might be None.
        '''
        result = self._get_plot_dims_attr_value('t')
        if result is not None:
            return result
        x, y, t = self.infer_xyt_dims(xy_fail_ok=True, t_fail_ok=True)
        if t is None:
            t_necessary_if = self._get_plot_dims_attr_value('t_necessary_if', default=lambda: True)
            if t_necessary_if():
                tattr = self.plot_dims_attrs['t']
                errmsg = f'failed to infer t_plot_dim (self.{tattr}) from dims={self.array.dims!r}, x={x!r}, y={y!r}'
                raise PlottingAmbiguityError(errmsg)
        return t
    @t_plot_dim.setter
    def t_plot_dim(self):
        self._set_plot_dims_attr_value('t', value)

    def infer_xyt_dims(self, *, xy_fail_ok=False, t_fail_ok=True):
        '''infers (x,y,t) dims from self.x, y, dim.
        xy_fail_ok, t_fail_ok: bool
            controls behavior when failing to infer a dim:
            True --> return None for that dim
            False --> raise PlottingAmbiguityError
            t_fail_ok corresponds to t; xy_fail_ok corresponds to x,y.

        see also: DEFAULTS.PLOT.DIMS_INFER, plotting.infer_xyt_dims, infer_xy_dims, infer_movie_dim
        '''
        x = self._get_plot_dims_attr_value('x')
        y = self._get_plot_dims_attr_value('y')
        t = self._get_plot_dims_attr_value('t')
        return infer_xyt_dims(self.array.dims, x=x, y=y, t=t, xy_fail_ok=xy_fail_ok, t_fail_ok=t_fail_ok)

    xy_plot_dims = simple_tuple_property('x_plot_dim', 'y_plot_dim',
                                         doc='''(self.x_plot_dim, self.y_plot_dim)''')
    xyt_plot_dims = simple_tuple_property('x_plot_dim', 'y_plot_dim', 't_plot_dim',
                                          doc='''(self.x_plot_dim, self.y_plot_dim, self.t_plot_dim)''')



"""
Package Purpose: common patterns for quantities. E.g. stats, arithmetic, derivatives.
"""
# import loaders
from .parenthesis_loader import ParenthesisLoader
from .basic_arithmetic import BasicArithmeticLoader
from .basic_derivatives import BasicDerivativeLoader
from .blur_loader import BlurLoader
from .caches_loader import CachesLoader
from .comparison_loader import ComparisonLoader
from .fft_loader import FFTLoader
from .fluids_loader import FluidsLoader
from .polyfit_loader import PolyfitLoader
from .stats_loader import StatsLoader
from .vector_arithmetic import VectorArithmeticLoader
from .vector_derivatives import VectorDerivativeLoader

# import other things
from .vector_arithmetic import (
    magnitude, unit_vector, angle_xy, angle_xy_to_hat,
    dot_product, cross_product, cross_component,
    take_perp_to, take_parallel_to,
)


"""
Package Purpose: PlasmaCalculator for different types of inputs
"""

from .bifrost import BifrostCalculator, BifrostMultifluidCalculator
from .ebysus import EbysusCalculator
from .eppic import (
    EppicCalculator, EppicHybridCalculator,
    EppicInstabilityCalculator, EppicTfbiCalculator, EppicCalculatorWithTfbi,
)
from .muram import MuramCalculator, MuramMultifluidCalculator
from .copapic import CopapicCalculator
